package addon;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

import cs.Azuma;
import robocode.Condition;
import robocode.ScannedRobotEvent;
import robocode.util.Utils;

/**
 * Tweaked to fit into my framework
 */
public class RaikoGun {
	private static double bearingDirection = 1, lastLatVel, lastVelocity,
	enemyEnergy, enemyDistance, lastVChangeTime, enemyLatVel,
	enemyVelocity;
	private static Point2D.Double enemyLocation;
	private static final int GF_ZERO = 15;
	private static final int GF_ONE = 30;
	private static String enemyName;
	private static int[][][][][][] guessFactors = new int[3][5][3][3][8][GF_ONE + 1];

	public Azuma bot;

	public RaikoGun(Azuma azuma) {
		bot = azuma;
	}

	public void onScannedRobot(ScannedRobotEvent e) {
		/*-------- setup data -----*/
		if (enemyName == null) {

			enemyName = e.getName();
			// restoreData();
		}

		Point2D.Double robotLocation = new Point2D.Double(bot.status.getX(),
				bot.status.getY());
		double theta;
		double enemyAbsoluteBearing = bot.status.getHeadingRadians()
		+ e.getBearingRadians();
		enemyDistance = e.getDistance();
		enemyLocation = projectMotion(robotLocation, enemyAbsoluteBearing, enemyDistance);

		enemyEnergy = e.getEnergy();

		Rectangle2D.Double BF = new Rectangle2D.Double(18, 18, 764, 564);

		MicroWave w = new MicroWave();

		lastLatVel = enemyLatVel;
		lastVelocity = enemyVelocity;
		enemyLatVel = (enemyVelocity = e.getVelocity())
		* Math.sin(e.getHeadingRadians() - enemyAbsoluteBearing);

		int distanceIndex = (int) enemyDistance / 140;

		double bulletPower = distanceIndex == 0 ? 3 : 2;
		theta = Math.min(bot.status.getEnergy() / 4,
				Math.min(enemyEnergy / 4, bulletPower));
		if (theta == bulletPower)
			bot.addCustomEvent(w);
		bulletPower = theta;
		w.bulletVelocity = 20D - 3D * bulletPower;

		int accelIndex = (int) Math.round(Math.abs(enemyLatVel)
				- Math.abs(lastLatVel));

		if (enemyLatVel != 0)
			bearingDirection = enemyLatVel > 0 ? 1 : -1;
			w.bearingDirection = bearingDirection
			* Math.asin(8D / w.bulletVelocity) / GF_ZERO;

			double moveTime = w.bulletVelocity * lastVChangeTime++ / enemyDistance;
			int bestGF = moveTime < .1 ? 1 : moveTime < .3 ? 2 : moveTime < 1 ? 3
					: 4;

			int vIndex = (int) Math.abs(enemyLatVel / 3);

			if (Math.abs(Math.abs(enemyVelocity) - Math.abs(lastVelocity)) > .6) {
				lastVChangeTime = 0;
				bestGF = 0;

				accelIndex = (int) Math.round(Math.abs(enemyVelocity)
						- Math.abs(lastVelocity));
				vIndex = (int) Math.abs(enemyVelocity / 3);
			}

			if (accelIndex != 0)
				accelIndex = accelIndex > 0 ? 1 : 2;

				w.firePosition = robotLocation;
				w.enemyAbsBearing = enemyAbsoluteBearing;
				// now using PEZ' near-wall segment
				w.waveGuessFactors = guessFactors[accelIndex][bestGF][vIndex][BF.contains(
						projectMotion(robotLocation, enemyAbsoluteBearing
								+ w.bearingDirection * GF_ZERO, enemyDistance)) ? 0
										: BF.contains(projectMotion(robotLocation, enemyAbsoluteBearing
												+ .5 * w.bearingDirection * GF_ZERO, enemyDistance)) ? 1
														: 2][distanceIndex];

				bestGF = GF_ZERO;

				for (int gf = GF_ONE; gf >= 0 && enemyEnergy > 0; gf--)
					if (w.waveGuessFactors[gf] > w.waveGuessFactors[bestGF])
						bestGF = gf;

				bot.setTurnGun(Utils.normalRelativeAngle(enemyAbsoluteBearing
						- bot.status.getGunHeadingRadians() + w.bearingDirection
						* (bestGF - GF_ZERO)));

				if (bot.status.getEnergy() > 1 || distanceIndex == 0)
					bot.setFire(bulletPower);
	}

	private static Point2D.Double projectMotion(Point2D.Double loc,
			double heading, double distance) {
		return new Point2D.Double(loc.x + distance * Math.sin(heading), loc.y
				+ distance * Math.cos(heading));
	}

	private static double absoluteBearing(Point2D.Double source,
			Point2D.Double target) {
		return Math.atan2(target.x - source.x, target.y - source.y);
	}

	class MicroWave extends Condition {

		Point2D.Double firePosition;
		int[] waveGuessFactors;
		double enemyAbsBearing, distance, bearingDirection, bulletVelocity;

		@Override
		public boolean test() {

			if ((RaikoGun.enemyLocation).distance(firePosition) <= (distance += bulletVelocity)
					+ bulletVelocity) {
				try {
					waveGuessFactors[(int) Math.round((Utils
							.normalRelativeAngle(absoluteBearing(firePosition,
									RaikoGun.enemyLocation) - enemyAbsBearing))
									/ bearingDirection + GF_ZERO)]++;
				} catch (ArrayIndexOutOfBoundsException e) {
				}
				bot.removeCustomEvent(this);
			}
			return false;
		}
	}
}
