/*
 * Copyright (c) 2011 Andrew Wells. All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification, are
 *	permitted provided that the following conditions are met:
 *
 *		1. Redistributions of source code must retain the above copyright notice, this list of
 *     conditions and the following disclaimer.
 *
 *		2. Redistributions in binary form must reproduce the above copyright notice, this list
 *    of conditions and the following disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 *	THIS SOFTWARE IS PROVIDED BY Andrew Wells ''AS IS'' AND ANY EXPRESS OR IMPLIED
 *	WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *	FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Andrew Wells OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *	ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *	NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * 	The views and conclusions contained in the software and documentation are those of the
 *	authors and should not be interpreted as representing official policies, either expressed
 *	or implied, of Andrew Wells.
 */

package aw.Mallorn.tree;

 /* Notes:
 * 		This tree allows weighting by allowing a bounding box that is not a hyper-cube,
 * but it does not allow weighting to be changed dynamically.
 * 
 *  	This tree does not cycle the dividing dimensions, instead, it uses the
 *  dimension in which the points have the widest variance.  This is good for 
 *  data that is not distributed evenly, but poor for data that is.
 *  
 *  	This tree is not intended to be constructed from previously known points.
 * 
 */

public class KDTreeBucket {
	public boolean isTerminal = true;
	public int dimensions;
	public int maxSize;
	int currentLastPoint = 0;
	double[][] points;
	public KDTreeBucket LeftBucket, RightBucket;
	public KDTreeBucket Parent;
	public int dividingDimension;
	public double divider;
	public double minimumBoundingRectangle[][];

	// it could be faster to not have this and check if we are in the correct
	// node instead.
	public NodesVisited nodesVisited = NodesVisited.None;

	public KDTreeBucket(int maxSize, int dimensions) {
		this.dimensions = dimensions;
		this.maxSize = maxSize;
		this.Parent = null;
		points = new double[maxSize][dimensions];
		this.minimumBoundingRectangle = new double[dimensions][2];
		for (int i = 0; i < dimensions; i++) {
			minimumBoundingRectangle[i][0] = Double.POSITIVE_INFINITY;
			minimumBoundingRectangle[i][1] = Double.NEGATIVE_INFINITY;
		}
	}

	public KDTreeBucket(KDTreeBucket Parent) {
		this.Parent = Parent;
		this.dimensions = Parent.dimensions;
		this.maxSize = Parent.maxSize;
		this.points = new double[maxSize][dimensions];
		this.minimumBoundingRectangle = new double[dimensions][2];
		for (int i = 0; i < dimensions; i++) {
			minimumBoundingRectangle[i][0] = Double.POSITIVE_INFINITY;
			minimumBoundingRectangle[i][1] = Double.NEGATIVE_INFINITY;
		}
	}
}
