/*
 * Copyright (c) 2011 Andrew Wells. All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification, are
 *	permitted provided that the following conditions are met:
 *
 *		1. Redistributions of source code must retain the above copyright notice, this list of
 *     conditions and the following disclaimer.
 *
 *		2. Redistributions in binary form must reproduce the above copyright notice, this list
 *    of conditions and the following disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 *	THIS SOFTWARE IS PROVIDED BY Andrew Wells ''AS IS'' AND ANY EXPRESS OR IMPLIED
 *	WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *	FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Andrew Wells OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *	ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *	NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * 	The views and conclusions contained in the software and documentation are those of the
 *	authors and should not be interpreted as representing official policies, either expressed
 *	or implied, of Andrew Wells.
 */

package aw.movement;

import aw.utils.ClassificationWeightingScheme;
import aw.utils.RoboGeom;
import aw.waves.MovementDataWave;
import aw.waves.DataWavePassedRecord;

import java.awt.Graphics2D;
import java.util.ArrayList;
import java.util.HashMap;

import robocode.util.Utils;

public class VirtualMovementsClassifier implements MovementClassifier {

	private ArrayList<MovementClassifier> Classifiers = new ArrayList<MovementClassifier>();
	private ArrayList<Integer> VirtualWavesClassifierIndices = new ArrayList<Integer>();

	// private ArrayList<Double> fastClassifierScores = new ArrayList<Double>();
	// private ArrayList<Double> mediumClassifierScores = new
	// ArrayList<Double>();
	// private ArrayList<Double> slowClassifierScores = new ArrayList<Double>();
	// private HashMap<MovementDataWave, ArrayList<double[][]>>
	// WavesAndClassifierPredictions = new HashMap<MovementDataWave,
	// ArrayList<double[][]>>();

	VirtualMovementsClassifier() {
		ClassificationWeightingScheme CWS = new AntiSimpleTargeterWeightingScheme();
		MovementClassifier MC = new KNNMovementClassifier(CWS, 36, 0, 5, 3,
				1.0 / 1, 0.0, 0.15);
		Classifiers.add(MC);

		CWS = new AntiSimpleTargeter2WeightingScheme();
		MC = new KNNMovementClassifier(CWS, 36, 0, 5, 3, 1.0 / 1, 0.0, 0.15);
		Classifiers.add(MC);

		CWS = new AntiSimpleTargeter3WeightingScheme();
		MC = new KNNMovementClassifier(CWS, 36, 0, 5, 3, 1.0 / 1, 0.0, 0.15);
		Classifiers.add(MC);

		MC = new AntiStandardTargeterClassifier();
		Classifiers.add(MC);

		MC = new AntiSemiAdvancedTargeterClassifier();
		Classifiers.add(MC);
		MC = new AntiSemiAdvancedTargeter2Classifier();
		Classifiers.add(MC);
		MC = new AntiSemiAdvancedTargeter3Classifier();
		Classifiers.add(MC);
		MC = new AntiSemiAdvancedTargeter4Classifier();
		Classifiers.add(MC);

		MC = new AntiAdvancedTargeterClassifier();
		Classifiers.add(MC);
		VirtualWavesClassifierIndices.add(8);

		MC = new AntiAdvancedTargeter2Classifier();
		Classifiers.add(MC);
		VirtualWavesClassifierIndices.add(9);
	}

	@Override
	public void train(MovementDataWave wave, DataWavePassedRecord rec) {
		for (MovementClassifier MC : Classifiers) {
			MC.train(wave, rec);
		}

		// if (WavesAndClassifierPredictions.containsKey(wave)) {
		// double angle = 0.5 * RoboGeom.preciseBotWidthAngle(
		// wave.getSourcePosition(),
		// wave.getBulletVelocity() * wave.getBulletTravelTime(),
		// rec.getHitAngle());
		//
		// // NOTE: this will not adjust for recently removed classifiers
		// ArrayList<double[][]> predictions = WavesAndClassifierPredictions
		// .get(wave);
		// for (int i = 0; i < predictions.size()
		// && i < fastClassifierScores.size(); i++) {
		//
		// double[][] AWBArray = predictions.get(i);
		// wave.setAnglesWeightsAndBandwidths(AWBArray);
		// double totalArea = Double.MIN_VALUE;
		// // for (double[] AWB : AWBArray) {
		// // totalArea += AWB[1];
		// // }
		// if (wave.getOrbitDir() == 1) {
		// totalArea = wave
		// .getDangerWithoutBulletShadowsForAngles(
		// wave.getGFOneAngle(),
		// wave.getGFNegOneAngle());
		// } else {
		// totalArea = wave
		// .getDangerWithoutBulletShadowsForAngles(
		// wave.getGFNegOneAngle(),
		// wave.getGFOneAngle());
		// }
		// double additionalScore = wave
		// .getDangerWithoutBulletShadowsForAngles(
		// rec.getHitAngle() + angle, rec.getHitAngle()
		// - angle)
		// / totalArea;
		// double newValue = fastClassifierScores.get(i);
		// newValue *= 0.7;
		// newValue += additionalScore;
		// fastClassifierScores.set(i, newValue);
		// newValue = mediumClassifierScores.get(i);
		// newValue *= 0.9;
		// newValue += additionalScore;
		// mediumClassifierScores.set(i, newValue);
		// newValue = slowClassifierScores.get(i);
		// newValue *= 0.97;
		// newValue += additionalScore;
		// slowClassifierScores.set(i, newValue);
		// }
		// }
	}

	// TODO make sure we don't change waves from the flattener after they have
	// been fired
	public void trainVirtualWave(MovementDataWave wave, DataWavePassedRecord rec) {
		// System.out.println("Start V Adding Points");

		for (Integer i : VirtualWavesClassifierIndices) {
			Classifiers.get(i).trainVirtualWave(wave, rec);
		}
		// System.out.println("Done V Adding Points");

	}

	public void trainSafelyPassedOfBullet(MovementDataWave wave) {
		// if (WavesAndClassifierPredictions.containsKey(wave)) {
		// // NOTE: this will not adjust for recently removed classifiers
		//
		// // TODO use large bulletShadows to train classifiers.
		// ArrayList<double[][]> predictions = WavesAndClassifierPredictions
		// .get(wave);
		// for (int i = 0; i < predictions.size()
		// && i < fastClassifierScores.size(); i++) {
		//
		// double[][] AWBArray = predictions.get(i);
		// wave.setAnglesWeightsAndBandwidths(AWBArray);
		// double totalArea = Double.MIN_VALUE;
		// // for (double[] AWB : AWBArray) {
		// // totalArea += AWB[1];
		// // }
		// if (wave.getOrbitDir() == 1) {
		// totalArea = wave
		// .getDangerWithoutBulletShadowsForAngles(
		// wave.getGFOneAngle(),
		// wave.getGFNegOneAngle());
		// } else {
		// totalArea = wave
		// .getDangerWithoutBulletShadowsForAngles(
		// wave.getGFNegOneAngle(),
		// wave.getGFOneAngle());
		// }
		// double additionalScore = 1.00
		// * wave.getDangerWithoutBulletShadowsForAngles(
		// wave.getMaxClockwiseAngle(),
		// wave.getMaxCounterClockwiseAngle()) / totalArea;
		// double newValue = fastClassifierScores.get(i);
		// newValue -= additionalScore;
		// fastClassifierScores.set(i, newValue);
		// newValue = mediumClassifierScores.get(i);
		// newValue -= additionalScore;
		// mediumClassifierScores.set(i, newValue);
		// newValue = slowClassifierScores.get(i);
		// newValue -= additionalScore;
		// slowClassifierScores.set(i, newValue);
		// }
		// }
	}

	@Override
	public double[][] getRawAnglesWeightsAndBandwidths(MovementDataWave wave) {
		// TODO use the AnglesWeightAndBandwidths class ?
		double maxScore = Double.NEGATIVE_INFINITY;
		int indexMaxScore = 0;
		double secondBestScore = Double.NEGATIVE_INFINITY;
		int indexSecondBestScore = 0;

		ArrayList<double[][]> predictions = new ArrayList<double[][]>();

		// for (int i = 0; i < Classifiers.size(); i++) {
		// predictions.add(Classifiers.get(i)
		// .getRawAnglesWeightsAndBandwidths(wave));
		// double score = fastClassifierScores.get(i)
		// + mediumClassifierScores.get(i)
		// + slowClassifierScores.get(i);
		// // double score = slowClassifierScores.get(i);
		//
		// if (score > maxScore) {
		// maxScore = score;
		// indexMaxScore = i;
		// } else if (score > secondBestScore) {
		// secondBestScore = score;
		// indexSecondBestScore = i;
		// }
		// }
		// WavesAndClassifierPredictions.put(wave, predictions);

		double[][] AWB = new double[predictions.get(indexMaxScore).length
				+ predictions.get(indexSecondBestScore).length][3];
		System.arraycopy(predictions.get(indexMaxScore), 0, AWB, 0,
				predictions.get(indexMaxScore).length);
		System.arraycopy(predictions.get(indexSecondBestScore), 0, AWB,
				predictions.get(indexMaxScore).length,
				predictions.get(indexSecondBestScore).length);

		return AWB;
	}

	public double[][] getRawAnglesWeightsAndBandwidths(MovementDataWave wave,
			double enemyHitPercentage, double enemyNormalizedHitPercentage,
			double enemyShotsLogged) {
		// TODO use the AnglesWeightAndBandwidths class ?
		// double maxScore = Double.NEGATIVE_INFINITY;
		// int indexMaxScore = 0;
		// double secondBestScore = Double.NEGATIVE_INFINITY;
		// int indexSecondBestScore = 0;

		ArrayList<double[][]> predictions = new ArrayList<double[][]>();

		// for (int i = 0; i < Classifiers.size(); i++) {
		// predictions.add(Classifiers.get(i)
		// .getRawAnglesWeightsAndBandwidths(wave));
		// double score = fastClassifierScores.get(i)
		// + mediumClassifierScores.get(i)
		// + slowClassifierScores.get(i);
		// // double score = slowClassifierScores.get(i);
		//
		// if (score > maxScore) {
		// maxScore = score;
		// indexMaxScore = i;
		// } else if (score > secondBestScore) {
		// secondBestScore = score;
		// indexSecondBestScore = i;
		// }
		// }
		// WavesAndClassifierPredictions.put(wave, predictions);

		for (int i = 0; i < Classifiers.size(); i++) {
			predictions.add(Classifiers.get(i)
					.getRawAnglesWeightsAndBandwidths(wave));
		}

		if (enemyNormalizedHitPercentage < -0.06 || enemyShotsLogged * enemyHitPercentage < 3.0) {

			double[][] AWB = new double[predictions.get(0).length
					+ predictions.get(1).length + predictions.get(2).length][3];
			System.arraycopy(predictions.get(0), 0, AWB, 0,
					predictions.get(0).length);
			System.arraycopy(predictions.get(1), 0, AWB,
					predictions.get(0).length, predictions.get(1).length);
			System.arraycopy(predictions.get(2), 0, AWB,
					predictions.get(0).length + predictions.get(1).length,
					predictions.get(2).length);
			return AWB;
		} else if (!flattenerEnabled(enemyHitPercentage,
				enemyNormalizedHitPercentage, enemyShotsLogged)) {
			if (enemyNormalizedHitPercentage < -0.035)
				return predictions.get(3);
			else {

				double[][] AWB = new double[predictions.get(0).length
						+ predictions.get(1).length + predictions.get(2).length
						+ predictions.get(3).length + predictions.get(4).length
						+ predictions.get(5).length + predictions.get(6).length
						+ predictions.get(7).length][3];

				System.arraycopy(predictions.get(0), 0, AWB, 0,
						predictions.get(0).length);
				System.arraycopy(predictions.get(1), 0, AWB,
						predictions.get(0).length, predictions.get(1).length);
				System.arraycopy(predictions.get(2), 0, AWB,
						predictions.get(0).length + predictions.get(1).length,
						predictions.get(2).length);

				System.arraycopy(predictions.get(3), 0, AWB,
						predictions.get(0).length + predictions.get(1).length
								+ predictions.get(2).length,
						predictions.get(3).length);
				
				System.arraycopy(predictions.get(4), 0, AWB,
						predictions.get(0).length + predictions.get(1).length
								+ predictions.get(2).length + predictions.get(3).length,
						predictions.get(4).length);
				
				System.arraycopy(predictions.get(5), 0, AWB,
						predictions.get(0).length + predictions.get(1).length
								+ predictions.get(2).length + predictions.get(3).length + predictions.get(4).length,
						predictions.get(5).length);
				
				System.arraycopy(predictions.get(6), 0, AWB,
						predictions.get(0).length + predictions.get(1).length
								+ predictions.get(2).length + predictions.get(3).length + predictions.get(4).length
								+ predictions.get(5).length,
						predictions.get(6).length);
				
				System.arraycopy(predictions.get(7), 0, AWB,
						predictions.get(0).length + predictions.get(1).length
						+ predictions.get(2).length + predictions.get(3).length + predictions.get(4).length
						+ predictions.get(5).length + predictions.get(6).length,
						predictions.get(7).length);
				
				
				return AWB;
			}

		} else {
			// double score2 = fastClassifierScores.get(2)
			// + mediumClassifierScores.get(2)
			// + slowClassifierScores.get(2);
			// double score3 = fastClassifierScores.get(3)
			// + mediumClassifierScores.get(3)
			// + slowClassifierScores.get(3);
			// if(score2 > score3)
			// return predictions.get(2);
			// else
			// return predictions.get(3);
			// if(score2 > score3) {
			// double[][] AWB = new double[predictions.get(1).length +
			// predictions.get(2).length][3];
			// System.arraycopy(predictions.get(1), 0, AWB, 0,
			// predictions.get(1).length);
			// System.arraycopy(predictions.get(2), 0, AWB,
			// predictions.get(1).length,
			// predictions.get(2).length);
			// return AWB;
			// }
			// else {
			// double[][] AWB = new double[predictions.get(1).length +
			// predictions.get(3).length][3];
			// System.arraycopy(predictions.get(1), 0, AWB, 0,
			// predictions.get(1).length);
			// System.arraycopy(predictions.get(3), 0, AWB,
			// predictions.get(1).length,
			// predictions.get(3).length);
			// return AWB;
			// }

			double[][] AWB = new double[predictions.get(3).length
					+ predictions.get(4).length + predictions.get(5).length
					+ predictions.get(6).length + predictions.get(7).length
					+ predictions.get(8).length + predictions.get(9).length][3];
			System.arraycopy(predictions.get(3), 0, AWB, 0,
					predictions.get(3).length);
			System.arraycopy(predictions.get(4), 0, AWB,
					predictions.get(3).length, predictions.get(4).length);
			System.arraycopy(predictions.get(5), 0, AWB,
					predictions.get(3).length + predictions.get(4).length,
					predictions.get(5).length);

			System.arraycopy(predictions.get(6), 0, AWB,
					predictions.get(3).length + predictions.get(4).length
							+ predictions.get(5).length,
					predictions.get(6).length);
			System.arraycopy(predictions.get(7), 0, AWB,
					predictions.get(3).length + predictions.get(4).length
							+ predictions.get(5).length
							+ predictions.get(6).length,
					predictions.get(7).length);
			System.arraycopy(predictions.get(8), 0, AWB,
					predictions.get(3).length + predictions.get(4).length
							+ predictions.get(5).length
							+ predictions.get(6).length
							+ predictions.get(7).length,
					predictions.get(8).length);
			System.arraycopy(predictions.get(9), 0, AWB,
					predictions.get(3).length + predictions.get(4).length
							+ predictions.get(5).length
							+ predictions.get(6).length
							+ predictions.get(7).length
							+ predictions.get(8).length,
					predictions.get(9).length);
			return AWB;
		}

		// double[][] AWB = new double[predictions.get(indexMaxScore).length
		// + predictions.get(indexSecondBestScore).length][3];
		// System.arraycopy(predictions.get(indexMaxScore), 0, AWB, 0,
		// predictions.get(indexMaxScore).length);
		// System.arraycopy(predictions.get(indexSecondBestScore), 0, AWB,
		// predictions.get(indexMaxScore).length,
		// predictions.get(indexSecondBestScore).length);
		//
		// return AWB;
	}

	/**
	 * This adds another MovementClassifier. It does not train this classifier
	 * with the previous situations.
	 * 
	 * @param classifier
	 *            The MovementClassifier to add.
	 */
	public void addClassifier(MovementClassifier classifier) {
		Classifiers.add(classifier);
	}

	public MovementClassifier getClassifier(int index) {
		return Classifiers.get(index);
	}

	public ArrayList<MovementClassifier> getClassifiers() {
		return Classifiers;
	}

	public boolean flattenerEnabled(double enemyHitPercentage,
			double enemyNormalizedHitPercentage, double enemyShotsLogged) {
		return (enemyShotsLogged * enemyHitPercentage >= 3.0)
				&& (enemyNormalizedHitPercentage >= 1.96 * Math
						.sqrt(enemyHitPercentage * (1.0 - enemyHitPercentage)
								/ enemyShotsLogged));
	}

	public void listClassifierScores() {
		// for (int i = 0; i < fastClassifierScores.size(); i++) {
		// double score = fastClassifierScores.get(i)
		// + mediumClassifierScores.get(i)
		// + slowClassifierScores.get(i);
		//
		// System.out.println("Movement classifier " + i + "(" + score + ")");
		// System.out.println("Movement classifier " + i + "fast score (" +
		// fastClassifierScores.get(i) + ")");
		// System.out.println("Movement classifier " + i + "medium score ("
		// + mediumClassifierScores.get(i) + ")");
		// System.out.println("Movement classifier " + i + "slow score ("
		// + slowClassifierScores.get(i) + ")\n");

		// }
	}

	// public boolean removeClassifier(MovementClassifier classifier) {
	// return Classifiers.remove(classifier);
	// }
}
