/*
 * Copyright (c) 2011 Andrew Wells. All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification, are
 *	permitted provided that the following conditions are met:
 *
 *		1. Redistributions of source code must retain the above copyright notice, this list of
 *     conditions and the following disclaimer.
 *
 *		2. Redistributions in binary form must reproduce the above copyright notice, this list
 *    of conditions and the following disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 *	THIS SOFTWARE IS PROVIDED BY Andrew Wells ''AS IS'' AND ANY EXPRESS OR IMPLIED
 *	WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *	FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Andrew Wells OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *	ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *	NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * 	The views and conclusions contained in the software and documentation are those of the
 *	authors and should not be interpreted as representing official policies, either expressed
 *	or implied, of Andrew Wells.
 */

package aw.waves;

import java.awt.geom.Point2D;

import robocode.Rules;
import robocode.util.Utils;

public class GunDataWave extends DataWave {
	private double virtuality;
	private double bulletPower;

	/**
	 * 
	 * @param time
	 *            Time to start wave
	 * @param x
	 *            X coordinate of wave center
	 * @param y
	 *            Y coordinate of wave center
	 * @param bulletPower
	 *            Wave bullet power
	 * @param dir
	 *            Direction to base GF's on (clockwise = 1, counter-clockwise =
	 *            -1)
	 * @param GFOne
	 *            Max absolute angle of GF 1
	 * @param GFNegOne
	 *            Max absolute angle of GF -1
	 * @param GFZero
	 *            GF 0
	 * @param dataPointCoordinates
	 *            coordinates for the k-dimensional data point
	 */
	public GunDataWave(long time, Point2D.Double sourcePosition,
			Point2D.Double targetPosition, double bulletPower,
			double bulletTravelTime, double Heading, double RelHeading,
			double Velocity, double AheadWallSeverity, double ReverseWallSeverity, double AheadWallDist, double ReverseWallDist,
			double Accel, double GFZeroAngle, double GFOneAngle,
			double GFNegOneAngle, long ticksSinceDirChange,
			long ticksSinceVelocityChange, long absFireTime,
			double virtuality, int orbitDir) {
		super(time, sourcePosition, targetPosition, Rules
				.getBulletSpeed(bulletPower), bulletTravelTime, Heading,
				RelHeading, Velocity, AheadWallSeverity, ReverseWallSeverity, AheadWallDist, ReverseWallDist, Accel, ticksSinceDirChange, ticksSinceVelocityChange, absFireTime,
				GFZeroAngle, GFOneAngle, GFNegOneAngle, orbitDir);

		this.virtuality = virtuality;
		this.bulletPower = bulletPower;
	}

	public double getVirtuality() {
		return virtuality;
	}

	public double getBulletPower() {
		return bulletPower;
	}
}
