package tjk.offense;

import robocode.*;
import robocode.util.Utils;
import tjk.universe.*;
import java.util.List;
import java.awt.Graphics2D;

/**
 * Gun - a class by tkiesel
 */
public class Gun
{
	
	private Universe universe;
	
	private Bot me;
	
	private AdvancedRobot robot;
	
	private Target target = null;
	
	public Gun(AdvancedRobot r, Universe u)
	{
		robot = r;
		universe = u;
		me = universe.getMe();
	}	
	
	/**
	 * onScannedRobot: What to do when you see another robot
	 */
	public void onScannedRobot(ScannedRobotEvent e) {
	    
		double firepower = 1.95;
		if ( e.getDistance() <= 55 )
		{
			firepower = 3.0;
		}
		else if ( e.getEnergy() <= 4 )
		{
			firepower = e.getEnergy()/4.0 + 0.001;
		}
		else if  ( e.getEnergy() < 6.0*firepower - 2.0  )
		{
			firepower = (e.getEnergy()+2.0)/6.0 + 0.001;
		}
		firepower = Math.min( me.getEnergy()-0.02 , firepower );
	
		if ( target != null && target.gunOnTarget(robot) && robot.getGunHeat() == 0.0000 && robot.getEnergy() > 0.1 ) 
		{
			robot.setFire(firepower);
			me.fireWave(firepower,universe.getTime());
			target = null;
		}
		else if ( universe.getRoundNum() <= 4 )
		{
			if ( universe.getTime() % 2 == 0 )
			{
				me.fireVWave(firepower,universe.getTime());	
			}
			else if ( universe.getTime() % 3 == 0 )
			{
				me.fireVWave(0.5,universe.getTime());	
			}
			else if ( universe.getTime() % 5 == 0 )
			{
				me.fireVWave(2.75,universe.getTime());	
			}
			else if ( universe.getTime() % 7 == 0 )
			{
				me.fireVWave(1.5,universe.getTime());	
			}
			else if ( universe.getTime() % 11 == 0 )
			{
				me.fireVWave(0.15,universe.getTime());	
			}
		}
	
	
		double turnsTillShoot = robot.getGunHeat() / robot.getGunCoolingRate();
		
		if ( turnsTillShoot > 5 )
		{
			target = null;
		}
		else
		{
		
			Wave w = me.getWave(true,firepower,universe.getTime());
			BracketHist h = me.getBracketHist(w, closestK() );
			target = getTarget(h,w);
		
		}
		
		
		if ( target != null && !target.isNull() )
		{
			robot.setTurnGunRightRadians( Utils.normalRelativeAngle( target.closestTargetHeading(robot.getGunHeadingRadians()) - robot.getGunHeadingRadians() ) );
		}
		else {
			robot.setTurnGunLeftRadians( Utils.normalRelativeAngle( robot.getGunHeadingRadians() - robot.getHeadingRadians() - e.getBearingRadians() ) );
		}
		
		/*
		if ( target != null && target.gunOnTarget(robot) && robot.getGunHeat() == 0.0000 && robot.getEnergy() > 0.1 ) 
		{
			robot.setFire(firepower);
			me.fireWave(firepower,universe.getTime());
			target = null;
		}
		*/
		
	}
	
	private int closestK()
	{
		return (int)Math.round(Math.min(Math.max(5,me.getStatSize()/10),150));
		/*
		if ( me.getStatSize() <= 30 )
		{
			return 5;
		}
		else if ( me.getStatSize() >= 1000 )
		{
			return 50;
		}
		return (int)Math.round( (135.0/970.0)*me.getStatSize() + (5.0 - 135.0/970.0) );
		*/
	}
	
	private Target getTarget( BracketHist h, Wave w )
	{
		List<double[]> l = h.maxBrackets();
		if ( l.isEmpty() )
		{
			System.out.println("FIXME: Gun trying to create empty Bracket.");
			return null;
		}
		return new Target(l, w);
	}
	
	public void paint(Graphics2D g)
	{
		if ( target != null )
		{
			target.paint(g, me);
		}
	}
	
}