package tjk.offense;

import robocode.*;
import robocode.util.Utils;
import tjk.universe.Wave;
import tjk.universe.Bot;
import java.util.ArrayList;
import java.util.List;
import java.awt.Color;
import java.awt.Graphics2D;

/**
 * Target - an array of TargetAngles (tkiesel)
 */
public class Target
{
	
	private ArrayList<TargetAngle> targetangles = null;
	
	// brackets numbers are least to greatest in guess-factor order.
	public Target( List<double[]> brackets, Wave w )
	{
		if ( brackets == null || brackets.isEmpty() )
		{
			targetangles = null;
			return;
		}
	
		for ( double[] d : brackets )
		{
			targetangles = new ArrayList<TargetAngle>(brackets.size());
			double a0 = Utils.normalAbsoluteAngle(w.GFToAbs(d[0]));
			double a1 = Utils.normalAbsoluteAngle(w.GFToAbs(d[1]));
			targetangles.add( new TargetAngle(a0,a1) );
		}
	}
	
	public boolean isNull()
	{
		return targetangles == null;
	}
	
	// Returns absolute target heading to a given heading.
	public double closestTargetHeading(double target)
	{
		if ( targetangles == null )
		{
			return 0.0;
		}
		double answer = Math.PI;
		for(TargetAngle t: targetangles)
		{
			double middle = middle(t.a0,t.a1);
			double delta = diff( middle , target);
			if ( Math.abs(delta) < Math.abs(answer) )
			{
				answer = middle;
			}
		}
		return answer;
	}
	
	public boolean gunOnTarget(AdvancedRobot r)
	{
		if ( targetangles == null )
		{
			return false;
		}
		double gun = r.getGunHeadingRadians();
		for (TargetAngle t: targetangles)
		{
			if ( absDiff(t.a0,gun) + absDiff(t.a1,gun) <= absDiff(t.a0,t.a1) + 0.01 )
			{
				return true;
			}
		}
		return false;
	}
	
	private static double absDiff(double a, double b)
	{
		return Math.abs(diff(a,b));
	}
	
	// diff between two absolute angles.. in relative angles.
	private static double diff(double a, double b)
	{
		double diff = a-b;
		while ( diff > Math.PI )
		{
			diff -= 2.0*Math.PI;
		}
		while ( diff < -Math.PI )
		{
			diff += 2.0*Math.PI;
		}
		return diff;
	}
	
	// center of two absolute angles, absolute angle
	private static double middle(double a, double b)
	{
		return Utils.normalAbsoluteAngle(a + 0.5*diff(b,a));
	}

	public void paint(Graphics2D g, Bot me)
	{
		g.setColor( new Color(255,0,0,100) );
		int rad = 1000;
		for (TargetAngle t: targetangles)
		{
			int start_angle = (int)Math.floor(Math.min(Math.toDegrees(t.a0), Math.toDegrees(t.a1) ) );
			int stop_angle = (int)Math.ceil( Math.abs(Math.toDegrees( diff(t.a1,t.a0) )) );
			g.fillArc( (int)me.getX()-rad, (int)me.getY()-rad, 2*rad, 2*rad, start_angle, stop_angle );
		}
	}
	
}
