package tjk.universe;

import robocode.*;
import robocode.util.Utils;
import tjk.utils.tjkUtil;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * Bot - a class by tkiesel
 */
public class Bot
{
	
	private String name = "";
	
	private Point2D.Double location,oldLocation;
	
	private double v,oldV,heading,energy,eDrop,heat,coolingRate,eCW;
	
	private boolean named = false;
	
	private boolean firedShot = false;
	
	private long zeroTime,vChangeTime;
	
	private ArrayList<Wave> waves = new ArrayList(5);
	
	private ArrayList<Wave> vWaves = new ArrayList(35);
	
	private Universe universe;
	
	private Bot enemy;
	
	private WaveStats stats;
	
	public Bot(Universe u) {
		universe = u;
		location = new Point2D.Double();
		oldLocation = new Point2D.Double();
		v = 0.0;
		oldV = 0.0;
		heading = 0.0;
		energy = 100.0;
		eDrop = 0.0;
		heat = 3.0;
		coolingRate = 0.1;
		eCW = 1.0;
		stats = new WaveStats();
		zeroTime = u.getTime();
		vChangeTime = u.getTime();
	}
	
	public void update(AdvancedRobot r, ScannedRobotEvent e) {
		if ( !named )
		{
			coolingRate = r.getGunCoolingRate();
			name = e.getName();
			named = true;
		}
		double x = r.getX() + e.getDistance()*Math.sin(r.getHeadingRadians() + e.getBearingRadians());
		double y = r.getY() + e.getDistance()*Math.cos(r.getHeadingRadians() + e.getBearingRadians());
		oldLocation.setLocation(location);
		location.setLocation(x,y);
		heading = e.getHeadingRadians();
		eDrop = energy-e.getEnergy();
		firedShot = checkFired( eDrop , v);
		if (firedShot)
		{
			heat = Rules.getGunHeat(eDrop);
		}
		if ( heat > 0.0 ) 
		{
			heat = Math.max(heat-coolingRate,0.0);
		}
		energy = e.getEnergy();
		oldV = v;
		v = e.getVelocity();
		if ( Math.abs(v - 0.0) <= 0.0001  )
		{
			zeroTime = getTime();
		}
		if ( Math.abs(v - oldV) >= 0.001 )
		{
			vChangeTime = getTime();
		}
		eCW = calcEnemyCW();
	}
	
	public void update(AdvancedRobot r) {
		if ( !named )
		{
			coolingRate = r.getGunCoolingRate();
			name = r.getName();
			named = true;
		}
		oldLocation.setLocation(location);
		location.setLocation(r.getX(),r.getY());
		heading = r.getHeadingRadians();
		eDrop = energy-r.getEnergy();
		firedShot = checkFired( eDrop , v );
		if (firedShot)
		{
			heat = Rules.getGunHeat(eDrop);
		}
		if ( heat > 0.0 ) 
		{
			heat = Math.max(heat-coolingRate,0.0);
		}
		energy = r.getEnergy();
		oldV = v;
		v = r.getVelocity();
		if ( Math.abs(v - 0.0) <= 0.0001  )
		{
			zeroTime = getTime();
		}
		if ( Math.abs(v - oldV) >= 0.001 )
		{
			vChangeTime = getTime();
		}
		eCW = calcEnemyCW();
		/*
		 * Heat sim is off until after first bullet of the round is fired. 
		 * Acceptable problem.
		 *
		if ( Math.abs(heat-r.getGunHeat()) > 0.05 )
		{
			System.out.println("Heat sim needs improvement! " + heat + " vs " + r.getGunHeat() );
		}
		*/
	}
	
	public void updateWaves()
	{	

		for ( Iterator<Wave> itr = vWaves.iterator(); itr.hasNext(); )
		{
			Wave w = itr.next();
			w.update();
			if ( w.getState() == Wave.DEAD )
			{
				// Load stale waves into WavePredictor
				stats.addWave(w);
				
				// Kill stale waves.
				itr.remove();
			}
		}
	
		for ( Iterator<Wave> itr = waves.iterator(); itr.hasNext(); )
		{
			Wave w = itr.next();
			w.update();
			if ( w.getState() == Wave.DEAD )
			{
				// Load stale waves into WaveStats
				stats.addWave(w);
				
				// Kill stale waves.
				itr.remove();
			}
		}
	
	}
	
	public void clearWaves()
	{
		waves.clear();
		vWaves.clear();
	}
	
	public BracketHist getBracketHist(Wave w, int closestK)
	{
		if ( w.getBracketHist() == null )
		{
			BracketHist bh = stats.getBracketHist(w,closestK);
			w.setBracketHist(bh);
			return bh;
		}
		return w.getBracketHist();
	}
	
	public Wave fireVWave(double pow, long time)
	{
		return fireWave(false,pow,time);
	}

	public Wave fireWave(double pow, long time)
	{
		return fireWave(true,pow,time);
	}
	
	public void checkFireWave()
	{
		if ( firedShot )
		{
			fireWave(true,eDrop,universe.getTime()-1);
		}
	}
	
	public Wave getWave(boolean real, double pow, long time)
	{
		return new Wave(this, enemy, time, pow, eCW, enemyMCM(Rules.getBulletSpeed(pow)), situation(real,pow));
	}
	
	private Wave fireWave(boolean real, double pow, long time)
	{
		Wave w = getWave(real,pow,time);
		
		if ( real )
		{
			waves.add( w );
		}
		else
		{
			vWaves.add( w );
		}
		return w;
	}
	
	public Iterator<Wave> getVWavesIterator()
	{
		return vWaves.iterator();
	}
	
	public Iterator<Wave> getWavesIterator()
	{
		return waves.iterator();
	}
	
	public boolean getFiredShot()
	{
		return firedShot;
	}
	
	public double bfH()
	{
		return universe.bfH();
	}

	public double bfW()
	{
		return universe.bfW();
	}
			
	public boolean checkFired(double eDrop, double V) {
		if ( eDrop >= 0.01 && eDrop <= 3.0 )
		{
			if ( Math.abs(Rules.getWallHitDamage(v)-eDrop) < 0.01 && ( location.getX() < 19 || location.getY() < 19 || location.getX() > universe.bfW()-19 || location.getY() > universe.bfH()-19 ) )
			{
				return false;
			}
			if ( Math.abs(Rules.ROBOT_HIT_DAMAGE-eDrop) < 0.01 && location.distanceSq(enemy.getLocation()) <= 2592  )
			{
				return false;
			}
			return true;
		}	
		return false;
	}
	
	public void setHeat(double h)
	{
		heat = h;
	}
	
	public void setEnemy(Bot e)
	{
		enemy = e;
	}
	
	public Bot getEnemy()
	{
		return enemy;
	}
	
	public String getName() {
		return name;
	}
	
	public Point2D.Double getLocation() {
		return location;
	}

	public Point2D.Double getOldLocation() {
		return oldLocation;
	}
				
	public double getX() {
		return location.getX();
	}
	
	public double getY() {
		return location.getY();
	}
	
	public double getV() {
		return v;
	}
	
	public double getOldV() {
		return oldV;
	}
	
	public double getA() {
		return v-oldV;
	}
	
	public int getStatSize()
	{
		return stats.getSize();
	}
	
	public double getHeading() {
		return heading;
	}

	public double headingToEnemy()
	{
		return Math.atan2( enemy.getX()-getX() , enemy.getY()-getY() );
	}

	public double enemyCW()
	{
		return eCW;
	}

	private double calcEnemyCW()
	{
		double v2 = enemy.getV();
		if ( v2 == 0.0 )
		{
			if ( enemy.getOldV() != 0.0 )
			{
				return -eCW;
			}
			return eCW;
		}
		double a = v2*Math.cos( enemy.getHeading() - headingToEnemy() - Math.PI/2.0 );
		if ( a >= 0 )
		{
			return 1.0;
		}
		else
		{
			return -1.0;
		}
	}
	
	public double closestLateralWallD()
	{
		return location.distance(closestLateralWallPoint());
	}
	
	public Point2D.Double closestLateralWallPoint()
	{
		double h = heading + Math.PI/2.0;
		double x0 = location.getX();
		double y0 = location.getY();
		
		// Check left wall.. x = 18
		double T = (17.5 - x0)/Math.sin(h);
		Point2D.Double answer = new Point2D.Double(x0+T*Math.sin(h), y0+T*Math.cos(h));
		double MinDist = location.distance(answer);
		
		// Check right wall.. x = universe.bfW()-18
		T = (universe.bfW() - 17.5 - x0)/Math.sin(h);
		Point2D.Double test = new Point2D.Double(x0+T*Math.sin(h), y0+T*Math.cos(h));
		double testD = location.distance(test);
		if ( testD < MinDist )
		{
			MinDist = testD;
			answer = test;
		}
		
		// Check bottom wall.. y = 18
		T = (17.5 - y0)/Math.cos(h);
		test = new Point2D.Double(x0+T*Math.sin(h), y0+T*Math.cos(h));
		testD = location.distance(test);
		if ( testD < MinDist )
		{
			MinDist = testD;
			answer = test;
		}
	
		// Check top wall.. y = universe.bfH() - 18
		T = (universe.bfH() - 17.5 - y0)/Math.cos(h);
		test = new Point2D.Double(x0+T*Math.sin(h), y0+T*Math.cos(h));
		testD = location.distance(test);
		if ( testD < MinDist )
		{
			MinDist = testD;
			answer = test;
		}
		
		return answer;
	}
	
	public double closestForwardWallD()
	{
		return location.distance(closestForwardWallPoint());
	}
	
	public Point2D.Double closestForwardWallPoint()
	{
		return closestWallPointInDirection(this.heading);
	}
	
	public double closestBackwardWallD()
	{
		return location.distance(closestBackwardWallPoint());
	}
	
	public Point2D.Double closestBackwardWallPoint()
	{
		return closestWallPointInDirection(this.heading+Math.PI);
	}
	
	public Point2D.Double closestWallPointInDirection(double h2)
	{
		double h = h2;
		if ( getV() < 0.0 )
		{
			h += Math.PI;
		}
		double x0 = location.getX();
		double y0 = location.getY();
		
		double T = 0.0;
		Point2D.Double answer = new Point2D.Double(-5000.0, -5000.0);
		double MinDist = Double.POSITIVE_INFINITY;
		Point2D.Double test = new Point2D.Double(-5000.0, -5000.0);
		double testD = Double.POSITIVE_INFINITY;
		
		// Check left wall.. x = 18
		T = (17.5 - x0)/Math.sin(h);
		if ( T >= 0 )
		{
			test = new Point2D.Double(x0+T*Math.sin(h), y0+T*Math.cos(h));
			testD = location.distance(test);
			if ( testD < MinDist )
			{
				answer = test;
				MinDist = testD;
			}
		}
		
		// Check right wall.. x = universe.bfW()-18
		T = (universe.bfW() - 17.5 - x0)/Math.sin(h);
		if ( T >= 0 )
		{
			test = new Point2D.Double(x0+T*Math.sin(h), y0+T*Math.cos(h));
			testD = location.distance(test);
			if ( testD < MinDist )
			{
				answer = test;
				MinDist = testD;
			}
		}
		
		// Check bottom wall.. y = 18
		T = (17.5 - y0)/Math.cos(h);
		if ( T >= 0 )
		{
			test = new Point2D.Double(x0+T*Math.sin(h), y0+T*Math.cos(h));
			testD = location.distance(test);
			if ( testD < MinDist )
			{
				answer = test;
				MinDist = testD;
			}
		}
	
		// Check top wall.. y = universe.bfH() - 18
		T = (universe.bfH() - 17.5 - y0)/Math.cos(h);
		if ( T >= 0 )
		{
			test = new Point2D.Double(x0+T*Math.sin(h), y0+T*Math.cos(h));
			testD = location.distance(test);
			if ( testD < MinDist )
			{
				answer = test;
				MinDist = testD;
			}
		}
		
		return answer;
	}
	
	public double[] enemyMCM(double vb)
	{
		//TODO: This is (slightly less) naive maxescapeangle.  Make it better with Sim.
		double heading = headingToEnemy();
		final double d0 = getLocation().distance(getEnemy().getLocation());
		final int vBot0 = (int)Math.round(Math.abs(getEnemy().getV()));
		
		int t = 0;
		double botD = 0.0;
		while ( vb*t*vb*t < d0*d0 + botD*botD )
		{
			botD = tjkUtil.distArrayForward[vBot0][t];
			t++;
		}
		double max = Utils.normalRelativeAngle(Math.atan(Math.abs(botD)/d0));
		
		t = 0;
		botD = 0.0;
		while ( vb*t*vb*t < d0*d0 + botD*botD )
		{
			botD = tjkUtil.distArrayBackward[vBot0][t];
			t++;
		}
		double min = Utils.normalRelativeAngle(Math.atan(Math.abs(botD)/d0));
		
		double[] r = {min,heading,max};
		return r;
	}

	public double[] situation(boolean real, double pow)
	{
		return Segmentation.situation(this,real,pow);
	}
	
	public long timeSinceVZero()
	{
		return getTime() - zeroTime;
	}

	public long timeSinceVChange()
	{
		return getTime() - vChangeTime;
	}
	
	public long getTime()
	{
		return universe.getTime();
	}
	
	public double getEnergy()
	{
		return energy;
	}
	
}