package tjk.universe;

import tjk.utils.KdTree.*;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Arrays;
/**
 * Bracket - a class by tkiesel
 */
public class BracketHist
{
	
	public ArrayList<Double> angles;
	public ArrayList<Integer> count; 
	
	public BracketHist( List<Entry<GFBracket>> brackets )
	{	
		angles = new ArrayList<Double>( 2*brackets.size() );
		count = new ArrayList<Integer>( 2*brackets.size() );
		double[] starts = new double[brackets.size()];
		double[] stops = new double[brackets.size()];
		
		Iterator<Entry<GFBracket>> itr = brackets.iterator();
		int i = 0;
		while ( itr.hasNext() )
		{
			Entry<GFBracket> e = itr.next();
			starts[i] = e.value.b[0];
			stops[i] = e.value.b[1];
			i++;
		}
		
		Arrays.sort(starts);
		Arrays.sort(stops);
		
		i = 0; // starts counter
		int j = 0; // stops counter
		int k = 0; // total counter
		int tempC = 0; // Current count.
		while ( k < 2*brackets.size() )
		{
			if ( i < starts.length && (j >= stops.length || Double.compare(starts[i],stops[j]) < 0) )
			{
				// If the next angle along the count is a start...
				tempC++;  // Up the count
				angles.add(new Double(starts[i])); // add the angle to the angles list.
				count.add(new Integer(tempC)); // add the count to the count list. 
				i++; // count off a start
				k++;
			}
			else if ( j < stops.length && (i >= starts.length || Double.compare(starts[i],stops[j]) > 0) )
			{
				// If the next angle along the count is an end...
				tempC--; // Drop the count
				angles.add(new Double(stops[j])); // add the angle to the angles list.
				count.add(new Integer(tempC)); // add the count to the count list.
				j++; // count off an end
				k++;
			}
			else if ( i < starts.length && j < stops.length && Double.compare(starts[i],stops[j]) == 0 )
			{
				// If a start and a stop happen coterminously, then nothing occurs at that angle!
				i++;
				j++;
				k += 2;
			}
			else
			{
				System.out.println("Not iterating through histogram properly!");
			}
			
		}
	
	}
	
	public List<double[]> maxBrackets()
	{
		List<double[]> answer = new ArrayList<double[]>();
		int max = -100;
		for (int i = 0; i < count.size(); i++)
		{
			int countHere = count.get(i).intValue();
			if ( countHere > max )
			{
				max = countHere;
				answer.clear();
				double[] newOne = {angles.get(i).doubleValue(),angles.get(i+1).doubleValue()};
				answer.add(newOne);
			}
			else if (countHere == max)
			{
				double[] newOne = {angles.get(i).doubleValue(),angles.get(i+1).doubleValue()};;
				answer.add(newOne);
			}
		}
		
		return answer;
	}
	
	/*
	public List<double[]> minBrackets()
	{
		List<double[]> answer = new ArrayList<double[]>();
		int min = angles.length + 10;
		for (int i = 0; i < angles.length-1; i++)
		{
			if ( count[i] < min )
			{
				min = count[i];
				answer.clear();
				double[] newOne = {angles[i],angles[i+1]};
				answer.add(newOne);
			}
			else if (count[i] == min)
			{
				double[] newOne = {angles[i],angles[i+1]};
				answer.add(newOne);
			}
		}
		
		return answer;
	}
	*/
	
}