package tjk.universe;

import robocode.Rules;

/**
 * Segmentation - a class by tkiesel
 */
public class Segmentation
{
	
	public static int DIMENSIONS = 10;
	public static int STATS_BUFFER = 5000;
	
	// real, latV, distance, accel, advancingV, timeSinceVzero, timeSinceVChange, forwardWallD, backwardWallD, lateralWallD
	public static double[] WEIGHTS = {3.0, 2.0, 1.0, 0.5, 1.0, 1.0, 1.5, 1.25, 1.0, 1.25};
	
	public static final double REAL = 0.0;
	public static final double VIRTUAL = 1.0;
	
	public static double[] situation(Bot b, boolean realWave, double power)
	{
		double real = REAL;
		if ( !realWave )
		{
			real = VIRTUAL;
		}
	
		double bulletV = 20.0 - 3.0*power;
	
		double[] sit = {real,latV(b),distance(b,bulletV),accel(b),adV(b),timeSinceVZero(b),timeSinceVChange(b),closestForwardWallD(b),closestBackwardWallD(b),closestLateralWallD(b)};
		return sit;
	}
	
	public static double distance(Bot b, double bulletV)
	{
		return ( b.getLocation().distance(b.getEnemy().getLocation())/bulletV  ) / (Universe.bfDiag()/11.0);
	}
	
	public static double closestForwardWallD(Bot b)
	{
		Bot e = b.getEnemy();
		return Math.pow( e.closestForwardWallD() / (Math.max(Universe.bfW(),Universe.bfH())/2.0) , 0.3333);
	}
	
	public static double closestBackwardWallD(Bot b)
	{
		Bot e = b.getEnemy();
		return Math.pow( e.closestBackwardWallD() / (Math.max(Universe.bfW(),Universe.bfH())/2.0) , 0.3333);
	}

	public static double closestLateralWallD(Bot b)
	{
		Bot e = b.getEnemy();
		return Math.pow( e.closestLateralWallD() / (Math.max(Universe.bfW(),Universe.bfH())/2.0) , 0.3333);
	}
	
	public static double accel(Bot b)
	{
		return Math.abs( b.getEnemy().getA() )/Rules.DECELERATION;
	}

	public static double latV(Bot b)
	{
		return Math.abs( b.getEnemy().getV()*Math.sin(b.getEnemy().getHeading()-b.getHeading()) )/Rules.MAX_VELOCITY;
	}
	
	public static double adV(Bot b)
	{
		return ( b.getEnemy().getV()*Math.cos(b.getEnemy().getHeading()-b.getHeading()) )/(2.0*Rules.MAX_VELOCITY)+0.5;
	}
	
	public static double timeSinceVZero(Bot b)
	{
		return Math.min(30.0,b.getEnemy().timeSinceVZero())/30.0;
	}
	
	public static double timeSinceVChange(Bot b)
	{
		return Math.min(30.0,b.getEnemy().timeSinceVChange())/30.0;
	}
	
	public static boolean isReal(double[] situation)
	{
		return situation[0] != VIRTUAL;
	}
	
}