package tjk.universe;

import robocode.*;
import java.awt.Graphics2D;
import java.awt.Color;

/**
 * Universe - a class by tkiesel
 */
public class Universe
{
	
	private AdvancedRobot robot;
	
	private Bot me,enemy;
	
	private static double battleFieldWidth,battleFieldHeight,battleFieldDiag;
	
	public Universe(AdvancedRobot r)
	{
		robot = r;
		me = new Bot(this);
		enemy = new Bot(this);
		me.setEnemy(enemy);
		enemy.setEnemy(me);
	}
	
	public void init()
	{
		me.setHeat(robot.getGunHeat());
		enemy.setHeat(robot.getGunHeat());
		battleFieldWidth = robot.getBattleFieldWidth();
		battleFieldHeight = robot.getBattleFieldHeight();
		battleFieldDiag = Math.sqrt(Math.pow(battleFieldWidth,2) + Math.pow(battleFieldHeight,2));
	}
	
	public void update(ScannedRobotEvent e) 
	{
		enemy.update(robot,e);
		me.update(robot);
		enemy.updateWaves();
		me.updateWaves();
		enemy.checkFireWave();
	}
	
	public void onDeath(DeathEvent e)
	{
		clearWaves();
	}
	
	public void onWin(WinEvent e)
	{
		clearWaves();
	}
	
	public void onRoundEnded(RoundEndedEvent e)
	{
		clearWaves();
		System.out.println("My wave stat size: " + getMe().getStatSize() + " (Offense)" );
		System.out.println("Enemy wave stat size: " + getEnemy().getStatSize() + " (Defense)" );
	}
	
	private void clearWaves()
	{
		me.clearWaves();
		enemy.clearWaves();
	}
	
	public void paint(Graphics2D g)
	{
		Painter.paintBot(g,me,new Color(0,255,0,55));
		//Painter.paintBot(g,enemy,new Color(255,0,0,55));
	}
	
	public Bot getMe()
	{
		return me;
	}

	public Bot getEnemy()
	{
		return enemy;
	}
	
	public static double bfW()
	{
		return battleFieldWidth;
	}

	public static double bfH()
	{
		return battleFieldHeight;
	}

	
	public static double bfDiag()
	{
		return battleFieldDiag;
	}	
	
	public long getTime()
	{
		return robot.getTime();
	}
	
	public int getRoundNum()
	{
		return robot.getRoundNum();
	}
	
}
