package tjk.universe;

import robocode.*;
import tjk.utils.tjkUtil;
import robocode.util.Utils;
import java.awt.geom.Point2D;

/**
 * Wave - a class by tkiesel
 */
public class Wave
{
	
	private Point2D.Double location;
	
	private Bot target;
	
	private double power,velocity,cW,radius;
	
	private double[] minCenterMax = {-0.1,0.0,0.1};
	
	private long launchtime,lifetime;
	
	private double[] hitBracket = {0.0,0.0};
	
	private double[] situation;
	
	private BracketHist stats = null;
	
	public static final byte ALIVE = 0;
	public static final byte HITTING = 1;
	public static final byte DEAD = 2;
	
	private byte state = ALIVE;
	
	public Wave(Bot source, Bot t, long time, double pow, double c, double[] mcm, double[] s)
	{
		target = t;
		
		if ( time == source.getTime()-1 )
		{
			Point2D.Double l = source.getOldLocation();
			location = new Point2D.Double(l.getX(),l.getY());
		}
		else {
			Point2D.Double l = source.getLocation();
			location = new Point2D.Double(l.getX(),l.getY());
		}
		launchtime = time;
		power = pow;
		velocity = Rules.getBulletSpeed(power);
		//velocity = 20.0 - 3.0*power;
		cW = source./*get*/enemyCW();
		minCenterMax = mcm;
		situation = s;
		radius = 0.0;
	}
	
	public BracketHist getBracketHist()
	{
		return stats;
	}

	public void setBracketHist(BracketHist bh)
	{
		stats = bh;
	}
	
	public byte getState()
	{
		return state;
	}
	
	public double getRadius()
	{
		return radius;
	}
	
	public double getX()
	{
		return location.getX();
	}

	public double getY()
	{
		return location.getY();
	}

	public double[] getMCM()
	{
		return minCenterMax;
	}

	public double[] getHitBracket()
	{
		return hitBracket;
	}

	public double[] getSituation()
	{
		return situation;
	}

	public double cW()
	{		
	     return cW;
	}

	public double GFToAbs(double i)
	{
		double angle = i;
		if ( angle < 0.0 )
		{
			angle = angle * minCenterMax[0];
		}
		else
		{
			angle = angle * minCenterMax[2];
		}
		return Utils.normalRelativeAngle(cW*angle + minCenterMax[1]);
	}

	public double absToGF(double i)
	{
		double angle = cW*Utils.normalRelativeAngle( i - minCenterMax[1] );
		if ( angle < 0.0 )
		{
			angle = angle / minCenterMax[0];
		}
		else
		{
			angle = angle / minCenterMax[2];
		}
		return angle;
	}

	public void addToBracket(double x,double y)
	{
		if ( x < -1 || y < -1 || x > target.bfW()+1 || y > target.bfH()+1 )
		{
			return;
		}
		double angle = absToGF(Math.atan2(x-location.getX(),y-location.getY()));
	
		if ( hitBracket[0] == 0.0 && hitBracket[1] == 0.0 )
		{
			hitBracket[0] =  angle;
			hitBracket[1] =  angle;
			//System.out.println(angle + " Guessfactor added to bracket.");
		}
		else if ( angle > hitBracket[1] )
		{
			hitBracket[1] = angle;
			//System.out.println(angle + " Guessfactor added to bracket.");
		}
		else if ( angle < hitBracket[0] )
		{
			hitBracket[0] = angle;
			//System.out.println(angle + " Guessfactor added to bracket.");
		}
	}

	// TODO:  Offload this code into a single object, or maybe Bot... this is too enormous to have
	//       on so many waves.
	public void update()
	{
		lifetime = target.getTime() - launchtime;
		radius = velocity*lifetime;
		double d = location.distance(target.getLocation());
		if ( state != DEAD && radius > d + 25.45585 + velocity )
		{
			state = DEAD;
		}
		else if (  state == ALIVE && radius >= d - 25.45558 )
		{
			state = HITTING;
		}
		// TODO: If hitting, add to the hitbrackets.
		if ( state != HITTING )
		{
			return;
		}
		// Bottom left corner
		double x_check = target.getX()-18.0;
		double y_check = target.getY()-18.0;
		double d_check = location.distance(x_check,y_check);
		if( d_check >= radius-velocity && d_check <= radius )
		{
			addToBracket(x_check,y_check);
		}
		// Top left corner
		x_check = target.getX()-18.0;
		y_check = target.getY()+18.0;
		d_check = location.distance(x_check,y_check);
		if( d_check >= radius-velocity && d_check <= radius )
		{
			addToBracket(x_check,y_check);
		}
		// Bottom right corner
		x_check = target.getX()+18.0;
		y_check = target.getY()-18.0;
		d_check = location.distance(x_check,y_check);
		if( d_check >= radius-velocity && d_check <= radius )
		{
			addToBracket(x_check,y_check);
		}
		// Top right corner
		x_check = target.getX()+18.0;
		y_check = target.getY()+18.0;
		d_check = location.distance(x_check,y_check);
		if( d_check >= radius-velocity && d_check <= radius )
		{
			addToBracket(x_check,y_check);
		}
		
		d_check = radius-velocity;
		
		//Left Side
		x_check = target.getX()-18.0;
		y_check = target.getY()-18.0;
		double x_check2 = target.getX()-18.0;
		double y_check2 = target.getY()+18.0;
		double[] sidecheck = tjkUtil.circleLineIntersect(x_check,y_check,x_check2,y_check2,location.getX(),location.getY(),radius);
		if ( sidecheck[0] > 0.0 )
		{
			addToBracket(sidecheck[1],sidecheck[2]);
			if ( sidecheck.length > 3 )
			{
				addToBracket(sidecheck[3],sidecheck[4]);
			}
		}
		sidecheck = tjkUtil.circleLineIntersect(x_check,y_check,x_check2,y_check2,location.getX(),location.getY(),d_check);
		if ( sidecheck[0] > 0.0 )
		{
			addToBracket(sidecheck[1],sidecheck[2]);
			if ( sidecheck.length > 3 )
			{
				addToBracket(sidecheck[3],sidecheck[4]);
			}
		}
		
		//Right Side
		x_check = target.getX()+18.0;
		y_check = target.getY()-18.0;
		x_check2 = target.getX()+18.0;
		y_check2 = target.getY()+18.0;
		sidecheck = tjkUtil.circleLineIntersect(x_check,y_check,x_check2,y_check2,location.getX(),location.getY(),radius);
		if ( sidecheck[0] > 0.0 )
		{
			addToBracket(sidecheck[1],sidecheck[2]);
			if ( sidecheck.length > 3 )
			{
				addToBracket(sidecheck[3],sidecheck[4]);
			}
		}
		sidecheck = tjkUtil.circleLineIntersect(x_check,y_check,x_check2,y_check2,location.getX(),location.getY(),d_check);
		if ( sidecheck[0] > 0.0 )
		{
			addToBracket(sidecheck[1],sidecheck[2]);
			if ( sidecheck.length > 3 )
			{
				addToBracket(sidecheck[3],sidecheck[4]);
			}
		}
		
		//Top
		x_check = target.getX()+18.0;
		y_check = target.getY()+18.0;
		x_check2 = target.getX()-18.0;
		y_check2 = target.getY()+18.0;
		sidecheck = tjkUtil.circleLineIntersect(x_check,y_check,x_check2,y_check2,location.getX(),location.getY(),radius);
		if ( sidecheck[0] > 0.0 )
		{
			addToBracket(sidecheck[1],sidecheck[2]);
			if ( sidecheck.length > 3 )
			{
				addToBracket(sidecheck[3],sidecheck[4]);
			}
		}
		sidecheck = tjkUtil.circleLineIntersect(x_check,y_check,x_check2,y_check2,location.getX(),location.getY(),d_check);
		if ( sidecheck[0] > 0.0 )
		{
			addToBracket(sidecheck[1],sidecheck[2]);
			if ( sidecheck.length > 3 )
			{
				addToBracket(sidecheck[3],sidecheck[4]);
			}
		}
		
		//Bottom
		x_check = target.getX()+18.0;
		y_check = target.getY()-18.0;
		x_check2 = target.getX()-18.0;
		y_check2 = target.getY()-18.0;
		sidecheck = tjkUtil.circleLineIntersect(x_check,y_check,x_check2,y_check2,location.getX(),location.getY(),radius);
		if ( sidecheck[0] > 0.0 )
		{
			addToBracket(sidecheck[1],sidecheck[2]);
			if ( sidecheck.length > 3 )
			{
				addToBracket(sidecheck[3],sidecheck[4]);
			}
		}
		sidecheck = tjkUtil.circleLineIntersect(x_check,y_check,x_check2,y_check2,location.getX(),location.getY(),d_check);
		if ( sidecheck[0] > 0.0 )
		{
			addToBracket(sidecheck[1],sidecheck[2]);
			if ( sidecheck.length > 3 )
			{
				addToBracket(sidecheck[3],sidecheck[4]);
			}
		}
		
		
	}
	
}