package voidious.gun;

import robocode.*;
import robocode.util.Utils;

import java.util.HashMap;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.Arrays;
import java.awt.Graphics2D;
import java.awt.geom.Point2D;
import java.awt.Color;

import voidious.Diamond;
import voidious.gun.EnemyGunData;
import voidious.gun.FireListener.DiaBullet;
import voidious.gfx.ColoredValueSet;
import voidious.gfx.RoboGraphic;
import voidious.gfx.RoboPainter;
import voidious.utils.BattleField;
import voidious.utils.DiaWave;
import voidious.utils.DiaUtils;
import voidious.utils.Predictor;
import voidious.utils.TimestampedFiringAngleSet;

/**
 * Copyright (c) 2009-2011 - Voidious
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 *    1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software.
 *
 *    2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 *    3. This notice may not be removed or altered from any source
 *    distribution.
 */

public class DiamondFist implements RoboPainter {
    private static final boolean ENABLE_DEBUGGING_GRAPHICS = true;
    private static final int WAVE_MATCH_THRESHOLD = 50;
    protected static final double NO_FIRING_ANGLE = -999;
    protected static final boolean AIMING = true;

    private static final double FAR_AWAY = 50000;
    private static final double DEFAULT_ENERGY = 100;
    private static final boolean IS_VISIT = true;
    private static final boolean IS_BULLET_HIT = false;
    private static final int INACTIVE_WAVE_OFFSET = -100;
    private static final DiaWave NO_WAVE_FOUND = null;
    private static final String NO_OPPONENT = "";

    private Diamond _robot;
    private boolean _tcMode;
    private Map<String, EnemyGunData> _enemies;
    private List<DiaWave> _waves;
    private VirtualGunsManager<TimestampedFiringAngleSet> _virtualGuns;
    private DuelGun<TimestampedFiringAngleSet> _currentGun;
    private MeleeGun _meleeGun;
    private String _opponentName;
    private int _enemiesAlive;
    private int _enemiesTotal;
    private long _lastBulletFiredTime;
    private double _aimedBulletPower;
    private FireListener _fireListener;
    private BattleField _battleField;

    private boolean _startedDuel;
    private Vector<RoboGraphic> _renderables;
    private boolean _painting;
    private boolean _robocodePainting;

    public DiamondFist(Diamond robot, boolean isTc) {
        _robot = robot;
        _tcMode = isTc;
        _enemies = new HashMap<String, EnemyGunData>();
        _waves = new ArrayList<DiaWave>();
        _enemiesTotal = _robot.getOthers();
        _battleField = new BattleField(_robot.getBattleFieldWidth(),
            _robot.getBattleFieldHeight());

        _opponentName = NO_OPPONENT;
        _fireListener = null;
        _renderables = new Vector<RoboGraphic>();
        _painting = false;
        _robocodePainting = false;

        initGuns();
    }

    private void initGuns() {
        _virtualGuns = new VirtualGunsManager<TimestampedFiringAngleSet>();

        DuelGun<TimestampedFiringAngleSet> mainGun;
        if (_enemiesTotal > 1) {
            mainGun = new MainGun(_enemies, _battleField, _renderables,
                _enemiesTotal);
        } else {
            mainGun = new TripHammerKnnGun(_enemies,_renderables);
        }
        _virtualGuns.addGun(mainGun);
        _currentGun = mainGun;

        DuelGun<TimestampedFiringAngleSet> antiSurfGun =
            new AntiSurferSimGun(_enemies, _battleField, _renderables,
                _enemiesTotal, _waves);
        _virtualGuns.addGun(antiSurfGun);

        _meleeGun = new MeleeGun(_enemies, _battleField, _renderables,
            _enemiesTotal);
    }

    private void initGunViews(EnemyGunData enemyData) {
        for (DuelGun<TimestampedFiringAngleSet> gun : _virtualGuns.getGuns()) {
            enemyData.registerDataViews(gun.newDataViews());
        }
        if (isMelee()) {
            enemyData.registerDataViews(_meleeGun.newDataViews());
        }
    }

    public void initRound(Diamond robot) {
        _robot = robot;
        for (EnemyGunData enemyData : _enemies.values()) {
            enemyData.energy = DEFAULT_ENERGY;
            enemyData.distance = FAR_AWAY;
            enemyData.alive = true;
            enemyData.pastLocations.clear();
        }
        _waves.clear();
        _enemiesAlive = _robot.getOthers();
        _lastBulletFiredTime = 0;
        _virtualGuns.clear();

        _startedDuel = false;
        _renderables.clear();
    }

    public void execute() {
        _enemiesAlive = _robot.getOthers();
        long currentTime = _robot.getTime();

        for (EnemyGunData enemyData : _enemies.values()) {
            if (enemyData.lastScanTime < currentTime) {
                enemyData.pastLocations.add(enemyData.location);
            }
        }

        updateBotDistancesSq();
        checkActiveWaves();
        if (is1v1()) {
            aimAndFire();
            if (!_startedDuel && !_opponentName.equals(NO_OPPONENT)) {
                _startedDuel = true;
                printCurrentGun();
            }
        } else {
            aimAndFireAtEveryone();
        }
    }

    private void aimAndFire() {
        EnemyGunData enemyData = closestLivingBot();
        if (enemyData != null) {
            fireIfGunTurned(_aimedBulletPower);

            DiaWave aimWave = enemyData.lastWaveFired;
            _aimedBulletPower = aimWave.bulletPower;
            Point2D.Double myNextLocation = Predictor.nextLocation(_robot);
            double firingAngle;
            if (enemyData.energy == 0 || ticksUntilGunCool() > 3) {
                firingAngle = DiaUtils.absoluteBearing(myNextLocation,
                    aimWave.targetLocation);
                evaluateVirtualGuns();
            } else {
                firingAngle = _currentGun.aimWithWave(aimWave, paintStatus());
            }
            _robot.setTurnGunRightRadians(Utils.normalRelativeAngle(
                firingAngle - _robot.getGunHeadingRadians()));
        }
    }

    private void aimAndFireAtEveryone() {
        EnemyGunData closestBot = closestLivingBot();
        if (closestBot != null) {
            fireIfGunTurned(_aimedBulletPower);
            Point2D.Double myNextLocation = Predictor.nextLocation(_robot);
            long ticksUntilFire = ticksUntilGunCool();

            if (ticksUntilFire % 2 == 0 || ticksUntilFire <= 4) {
                _aimedBulletPower = calculateBulletPower();
                double firingAngle = _meleeGun.aimAtEveryone(myNextLocation,
                        _robot.getTime(), _enemiesAlive, _aimedBulletPower,
                    closestBot, paintStatus());
                _robot.setTurnGunRightRadians(Utils.normalRelativeAngle(
                    firingAngle - _robot.getGunHeadingRadians()));
            }
        }
    }

    private Point2D.Double myLocation() {
        return new Point2D.Double(_robot.getX(), _robot.getY());
    }

    private double calculateBulletPower() {
        EnemyGunData enemyData = closestLivingBot();
        double bulletPower = 3;
        if (enemyData != null) {
            double myEnergy = _robot.getEnergy();
            if (_tcMode) {
                bulletPower = Math.min(myEnergy, 3);
            } else if (is1v1()) { 
                bulletPower = 1.95;

                if (enemyData.distance < 150) {
                    bulletPower = 2.999;
                }

                if (myEnergy < 35 && enemyData.distance > 120) {
                    bulletPower = Math.min(bulletPower, 
                        DiaUtils.cube(myEnergy / 35) * 1.95);
                }

                bulletPower = Math.min(bulletPower, enemyData.energy / 4);
                bulletPower = Math.max(bulletPower, 0.1);
                bulletPower = Math.min(bulletPower, myEnergy);
            } else {
                double avgEnemyEnergy = avgEnemyEnergy();

                bulletPower = 2.999;

                if (_enemiesAlive <= 3) {
                    bulletPower = 1.999;
                }

                if (_enemiesAlive <= 5 && enemyData.distance > 500) {
                    bulletPower = 1.499;
                }

                if ((myEnergy < avgEnemyEnergy && _enemiesAlive <= 5 && 
                    enemyData.distance > 300) || enemyData.distance > 700) {

                    bulletPower = 0.999;
                }

                if (myEnergy < 20 && myEnergy < avgEnemyEnergy) {
                    bulletPower = Math.min(bulletPower, 2 - 
                        ((20 - myEnergy) / 11));
                }

                bulletPower = Math.max(bulletPower, 0.1);
                bulletPower = Math.min(bulletPower, myEnergy);
            }
        }

        return bulletPower;
    }

    private void evaluateVirtualGuns() {
        DuelGun<TimestampedFiringAngleSet> bestGun =
            _virtualGuns.bestGun(_opponentName);
        if (_currentGun != bestGun) {
            _currentGun = bestGun;
            System.out.println("Switching to " + _currentGun.getLabel() + " (" +
                DiaUtils.round(
                    _virtualGuns.getRating(_currentGun, _opponentName) * 100, 2) 
                + ")");
        }
    }

    private void fireIfGunTurned(double bulletPower) {
        if (_robot.getGunHeat() == 0 && _robot.getGunTurnRemaining() == 0) {
            Bullet realBullet = null;
            if (_tcMode) {
                realBullet = _robot.setFireBullet(3);
            } else if (_robot.getEnergy() > bulletPower) {
                realBullet = setFireBulletLogged(bulletPower);
            }

            if (realBullet != null) {
                _lastBulletFiredTime = _robot.getTime();
                markFiringWaves();
            }
        }
    }

    public void onScannedRobot(ScannedRobotEvent e) {
        String botName = e.getName();
        Point2D.Double enemyLocation = DiaUtils.translateToField(
            DiaUtils.project(myLocation(), Utils.normalAbsoluteAngle(
                e.getBearingRadians() + _robot.getHeadingRadians()), 
            e.getDistance()), _battleField);

        EnemyGunData enemyData;
        if (_enemies.containsKey(botName)) {
            enemyData = _enemies.get(botName);
            enemyData.timeSinceDirectionChange += 
                e.getTime() - enemyData.lastScanTime;
            enemyData.timeSinceVelocityChange += 
                e.getTime() - enemyData.lastScanTime;

            if (Math.abs(e.getVelocity() - 
                    enemyData.lastNonZeroVelocity) > 0.5) {
                enemyData.timeSinceVelocityChange = 0;
            }

            if (Math.abs(e.getVelocity()) > 0.5) {
                if (Math.signum(e.getVelocity()) != 
                        Math.signum(enemyData.lastNonZeroVelocity)) {
                    enemyData.timeSinceDirectionChange = 0;
                }
                enemyData.lastNonZeroVelocity = e.getVelocity();
            }

            enemyData.distance = e.getDistance();
            enemyData.energy = e.getEnergy();
            enemyData.location = enemyLocation;
            enemyData.lastScanTime = e.getTime();
            enemyData.previousVelocity = enemyData.velocity;
            enemyData.velocity = e.getVelocity();
            enemyData.heading = e.getHeadingRadians();
        } else {
            enemyData = new EnemyGunData(e.getName(), e.getDistance(), 
                e.getEnergy(), enemyLocation, e.getTime(), e.getVelocity(), 
                e.getHeadingRadians());
            _enemies.put(botName, enemyData);
            initGunViews(enemyData);
        }
        enemyData.pastLocations.add(enemyLocation);

        fireNextTickWave(enemyLocation, botName, calculateBulletPower());

        if (is1v1()) {
            _opponentName = e.getName();
        }
    }

    public void onRobotDeath(RobotDeathEvent e) {
        try {
            _enemies.get(e.getName()).alive = false;
        } catch (java.lang.NullPointerException npe) {
            System.out.println(
                "WARNING (gun):  A bot died that I never knew existed!");
        }
    }

    public void onHitByBullet(HitByBulletEvent e) {
        try {
           _enemies.get(e.getName()).damageTaken += 
               Rules.getBulletDamage(e.getBullet().getPower());
        } catch (java.lang.NullPointerException npe) {
            System.out.println(
                "WARNING (gun):  A bot shot me that I never knew existed!");
        }
    }

    public void onWin(WinEvent e) {
        roundOver();
    }

    public void onDeath(DeathEvent e) {
        roundOver();
    }

    private void roundOver() {
        if (_tcMode) {
            System.out.println("TC score: " +
                damageGiven() / (_robot.getRoundNum() + 1));
        }
        if (is1v1()) {
            _virtualGuns.printGunRatings(_opponentName);
        }
    }

    public void onBulletHit(BulletHitEvent e) {
        String botName = e.getName();
        if (_enemies.containsKey(botName)) {
            double bulletDamage = Math.min(
                Rules.getBulletDamage(e.getBullet().getPower()),
                _enemies.get(botName).energy);
            _enemies.get(botName).damageGiven += bulletDamage;
            processBulletHit(e.getBullet(), botName);
        } else {        
            System.out.println(
                "WARNING (gun):  I shot a bot that I never knew existed!");
        }
    }

    public void onBulletHitBullet(BulletHitBulletEvent e) {
        String botName = e.getHitBullet().getName();
        if (_enemies.containsKey(botName)) {
            processBulletHit(e.getBullet(), botName);
        } else {        
            System.out.println("WARNING (gun):  I shot a bullet by a bot " +
                "that I never knew existed!");
        }
    }

    private void processBulletHit(Bullet bullet, String botName) {
        EnemyGunData enemyData = _enemies.get(botName);
        long currentTime = _robot.getTime();
        Point2D.Double bulletLocation =
            new Point2D.Double(bullet.getX(), bullet.getY());
        DiaWave hitWave = DiaWave.findClosestWave(_waves, bulletLocation,
            currentTime, DiaWave.FIRING_WAVE, WAVE_MATCH_THRESHOLD, botName,
            bullet.getPower());
        if (hitWave != NO_WAVE_FOUND) {
            Point2D.Double hitVector = 
                hitWave.displacementVector(enemyData.location, currentTime);
            double hitFactor = 0;
            if (_enemiesTotal == 1) {
                hitFactor = hitWave.guessFactorPrecise(enemyData.location);
            }
            enemyData.registerWave(hitWave, hitVector, hitFactor,
                _robot.getRoundNum(), _robot.getTime(), IS_BULLET_HIT);
            hitWave.processedBulletHit = true;
        }
    }

    public void onPaint(Graphics2D g) {
        if (paintStatus()) {
            for (RoboGraphic r : _renderables) {
                r.render(g);
            }
            _renderables.clear();
        }
    }

    private double damageGiven() {
        double damageGiven = 0;
        for (EnemyGunData enemyData : _enemies.values()) {
            damageGiven += enemyData.damageGiven;
        }
        return damageGiven;
    }

    private void fireNextTickWave(Point2D.Double targetLocation, 
            String targetName, double bulletPower) {
        EnemyGunData targetData = _enemies.get(targetName);
        Point2D.Double myNextLocation = Predictor.nextLocation(_robot);
        Point2D.Double enemyNextLocation = DiaUtils.translateToField(
            Predictor.nextLocation(targetLocation, targetData.velocity, 
                targetData.heading), _battleField);

        double accel = DiaUtils.limit(-Rules.DECELERATION, 
            DiaUtils.accel(targetData.velocity, targetData.previousVelocity), 
            Rules.ACCELERATION);
        double dl8t = targetLocation.distance(targetData.getPastLocation(8));
        double dl20t = targetLocation.distance(targetData.getPastLocation(20));
        double dl40t = targetLocation.distance(targetData.getPastLocation(40));

        DiaWave nextWave = new DiaWave(targetName, myNextLocation,
            enemyNextLocation, _robot.getTime() + 1, bulletPower, 
            targetData.heading, targetData.velocity,
            DiaUtils.nonZeroSign(targetData.lastNonZeroVelocity), _battleField)
                .setAccel(accel)
                .setDistance(myNextLocation.distance(enemyNextLocation))
                .setDchangeTime(targetData.timeSinceDirectionChange)
                .setVchangeTime(targetData.timeSinceVelocityChange)
                .setDistanceLast8Ticks(dl8t)
                .setDistanceLast20Ticks(dl20t)
                .setDistanceLast40Ticks(dl40t)
                .setTargetEnergy(targetData.energy)
                .setSourceEnergy(_robot.getEnergy())
                .setGunHeat(_robot.getGunHeat())
                .setEnemiesAlive(_enemiesAlive)
                .setLastBulletFiredTime(_lastBulletFiredTime);
        setWallDistances(nextWave);

        targetData.lastWaveFired = nextWave;
        if (_robot.getEnergy() > 0) {
            _waves.add(nextWave);
        }

        for (EnemyGunData enemyData : _enemies.values()) {
            if (enemyData.alive && !enemyData.botName.equals(targetName) &&
                    !enemyData.pastLocations.isEmpty()) {
                Point2D.Double altNextLocation = DiaUtils.translateToField( 
                    Predictor.nextLocation(enemyData.location, 
                        enemyData.velocity, enemyData.heading), _battleField);

                accel = DiaUtils.limit(-Rules.DECELERATION, 
                    DiaUtils.accel(enemyData.velocity,
                        enemyData.previousVelocity), 
                    Rules.ACCELERATION);
                dl8t = targetLocation.distance(enemyData.getPastLocation(8));
                dl20t = targetLocation.distance(enemyData.getPastLocation(20));
                dl40t = targetLocation.distance(enemyData.getPastLocation(40));

                DiaWave altWave = new DiaWave(targetName, altNextLocation,
                    enemyNextLocation, _robot.getTime() + 1,
                    bulletPower, targetData.heading, targetData.velocity, 
                    DiaUtils.nonZeroSign(targetData.lastNonZeroVelocity),
                    _battleField)
                        .setAccel(accel)
                        .setDistance(
                            altNextLocation.distance(enemyNextLocation))
                        .setDchangeTime(targetData.timeSinceDirectionChange)
                        .setVchangeTime(targetData.timeSinceVelocityChange)
                        .setDistanceLast8Ticks(dl8t)
                        .setDistanceLast20Ticks(dl20t)
                        .setDistanceLast40Ticks(dl40t)
                        .setTargetEnergy(enemyData.energy)
                        .setSourceEnergy(_robot.getEnergy())
                        .setGunHeat(_robot.getGunHeat())
                        .setEnemiesAlive(_enemiesAlive)
                        .setLastBulletFiredTime(_lastBulletFiredTime)
                        .setAltWave(true);
                setWallDistances(altWave);

                if (_robot.getEnergy() > 0) {
                    _waves.add(altWave);
                }
            }
        }
    }

    private void checkActiveWaves() {
        int currentRound = _robot.getRoundNum();
        long currentTime = _robot.getTime();

        Iterator<DiaWave> wavesIterator = _waves.iterator();
        while (wavesIterator.hasNext()) {
            DiaWave w = wavesIterator.next();
            EnemyGunData enemyData = _enemies.get(w.botName);

            if (w.fireTime == currentTime && !w.altWave) {
                w.gunHeat = _robot.getGunHeat();
                w.lastBulletFiredTime = _lastBulletFiredTime;
                boolean recalcWallDistance = false;
                double bulletPower = calculateBulletPower();
                if (w.bulletPower != bulletPower) {
                    w.setBulletPower(bulletPower);
                    recalcWallDistance = true;
                }

                if (enemyData.lastScanTime == w.fireTime) {
                    // If we get two scans in a row, we can overwrite our
                    // estimates with real data.
                    w.targetLocation = enemyData.location;
                    w.sourceLocation = myLocation();
                    w.absBearing = DiaUtils.absoluteBearing(
                        w.sourceLocation, w.targetLocation);

                    // Technically might recalc wall distance here, too,
                    // but lotsa CPU for negligible difference.
                }

                if (recalcWallDistance) {
                    w.clearCachedPreciseEscapeAngles();
                    _virtualGuns.clearWave(w);
                    setWallDistances(w);
                }
            }

            if (!enemyData.alive) {
                wavesIterator.remove();
            } else if (!w.processedWaveBreak && w.wavePassedInterpolate(
                    enemyData.location, enemyData.lastScanTime, currentTime)) {
                double guessFactor = 0;
                double[] preciseBotWidth = null;
                if (_enemiesTotal == 1 && !w.waveBreakStates.isEmpty()) {
                    preciseBotWidth = w.preciseBotWidth(w.waveBreakStates);
                    double hitAngle = preciseBotWidth[0];
                    guessFactor = w.guessFactorPrecise(hitAngle);
                }

                Point2D.Double dispVector = w.displacementVector();
                enemyData.registerWave(w, dispVector, guessFactor, currentRound,
                    currentTime, IS_VISIT);
                w.processedWaveBreak = true;

                if (w.enemiesAlive == 1 && w.firingWave && !w.altWave
                        && !w.waveBreakStates.isEmpty()) {
                    if (preciseBotWidth == null) {
                        preciseBotWidth = w.preciseBotWidth(w.waveBreakStates);
                    }
                    double hitAngle = preciseBotWidth[0];
                    double tolerance = preciseBotWidth[1];
                    _virtualGuns.registerWaveBreak(w, hitAngle, tolerance);
                }

                if (paintStatus() && isMelee()) {
                    drawDisplacementVector(w);
                }
            } else if (w.wavePassed(enemyData.location, currentTime, 
                    INACTIVE_WAVE_OFFSET)) {
                wavesIterator.remove();
            }
        }
    }

    private void markFiringWaves() {
        long currentTime = _robot.getTime();
        for (DiaWave w : _waves) {
            if (w.fireTime == currentTime && !w.altWave) {
                w.firingWave = true;
                if (is1v1()) {
                    _virtualGuns.fireVirtualBullets(w);
                }
            }
        }
    }

    private void setWallDistances(DiaWave w) {
        if (is1v1()) {
            w.targetWallDistance = Math.min(1.5,
                DiaUtils.orbitalWallDistance(w.sourceLocation, w.targetLocation,
                    w.bulletPower, w.orbitDirection, _battleField));
            w.targetRevWallDistance = Math.min(1.5,
                DiaUtils.orbitalWallDistance(w.sourceLocation, w.targetLocation,
                    w.bulletPower, -w.orbitDirection, _battleField));
        } else {
            double targetDistance = w.sourceLocation.distance(w.targetLocation);
            w.targetWallDistance = Math.min(1.5,
                DiaUtils.directToWallDistance(w.targetLocation,
                    targetDistance, w.effectiveHeading(), w.bulletPower,
                    _battleField));
            w.targetRevWallDistance = Math.min(1.5,
                DiaUtils.directToWallDistance(w.targetLocation, 
                    targetDistance, w.effectiveHeading() + Math.PI,
                    w.bulletPower, _battleField));
        }
    }

    private void updateBotDistancesSq() {
        if (_enemies.size() <= 1) { return; }

        String[] botNames = new String[_enemies.size()];
        _enemies.keySet().toArray(botNames);

        for (int x = 0; x < botNames.length; x++) {
            EnemyGunData enemyData1 = _enemies.get(botNames[x]);
            for (int y = x + 1; y < botNames.length; y++) {
                EnemyGunData enemyData2 = _enemies.get(botNames[y]);
                if (enemyData1.alive && enemyData2.alive) {
                    double distanceSq =
                        enemyData1.location.distanceSq(enemyData2.location);
                    enemyData1.setBotDistanceSq(botNames[y], distanceSq);
                    enemyData2.setBotDistanceSq(botNames[x], distanceSq);
                } else {
                    if (!enemyData1.alive) {
                        enemyData2.removeDistanceSq(botNames[x]);
                    }
                    if (!enemyData2.alive) {
                        enemyData1.removeDistanceSq(botNames[y]);
                    }
                }
            }
        }
    }

    private long ticksUntilGunCool() {
        return Math.round(Math.ceil(
            _robot.getGunHeat() / _robot.getGunCoolingRate()));
    }

    private double avgEnemyEnergy() {
        double totalEnergy = 0;
        for (EnemyGunData enemyData : _enemies.values()) {
            if (enemyData.alive) {
                totalEnergy += enemyData.energy;
            }
        }
        return totalEnergy / _enemiesAlive;
    }

    private EnemyGunData closestLivingBot() {
        EnemyGunData closestBot = null;
        double closestDistance = Double.POSITIVE_INFINITY;
        for (EnemyGunData enemyData : _enemies.values()) {
            if (enemyData.alive) {
                double thisDistance =
                    myLocation().distanceSq(enemyData.location);
                if (thisDistance < closestDistance) {
                    closestBot = enemyData;
                    closestDistance = thisDistance;
                }
            }
        }
        return closestBot;
    }

    private boolean is1v1() {
        return (_enemiesAlive <= 1);
    }

    private boolean isMelee() {
        return (_enemiesAlive > 1);
    }

    public int getEnemiesAlive() {
        return _enemiesAlive;
    }

    public void setFireListener(FireListener listener) {
        _fireListener = listener;
    }

    private Bullet setFireBulletLogged(double bulletPower) {
        Bullet bullet = _robot.setFireBullet(bulletPower);
        if (_fireListener != null && bullet != null) {
            _fireListener.bulletFired(new DiaBullet(
                _robot.getTime(), myLocation(), _robot.getGunHeadingRadians(),
                (20 - (3 * bulletPower))));
        }
        return bullet;
    }

    public void paintOn() {
        _painting = ENABLE_DEBUGGING_GRAPHICS;
    }

    public void paintOff() {
        _renderables.clear();
        _painting = false;
    }

    public void robocodePaintOn() {
        _robocodePainting = true;
    }

    public void robocodePaintOff() {
        _renderables.clear();
        _robocodePainting = false;
    }

    public String paintLabel() {
        return "Gun";
    }

    public boolean paintStatus() {
        return (_painting && _robocodePainting);
    }

    protected static void drawGunAngles(Vector<RoboGraphic> renderables,
        DiaWave wave, ColoredValueSet cvs, double bestAngle, double bandwidth) {

        double arrowLength = Math.min(250, 
            wave.sourceLocation.distance(wave.targetLocation) - 75);

        for (ColoredValueSet.ColoredValue cv : cvs.getColoredValues()) {
            Color c = cv.grayToWhiteColor();
            Point2D.Double angleHead = DiaUtils.project(wave.sourceLocation, 
                cv.firingAngle, arrowLength);

            renderables.add(RoboGraphic.drawLine(wave.sourceLocation,
                angleHead, c));
            renderables.addAll(Arrays.asList(RoboGraphic.drawArrowHead(
                angleHead, 10, cv.firingAngle, c)));
        }

        renderables.add(RoboGraphic.drawPoint(DiaUtils.project(
            wave.sourceLocation, bestAngle, arrowLength + 5), Color.red));
    }

    private void drawDisplacementVector(DiaWave w) {
        Point2D.Double dispVector = w.displacementVector();
        long currentTime = _robot.getTime();
        double drawBearing = (w.orbitDirection * 
            DiaUtils.absoluteBearing(DiaWave.ORIGIN, dispVector)) +
            w.effectiveHeading();
        double drawDistance = dispVector.distance(DiaWave.ORIGIN) *
            w.waveBreakBulletTicks();
        _renderables.add(RoboGraphic.drawLine(
            DiaUtils.project(w.targetLocation, w.targetHeading, 
                25),
            DiaUtils.project(w.targetLocation, 
                w.targetHeading + Math.PI, 25), 
            Color.darkGray));
        _renderables.add(RoboGraphic.drawLine(
            DiaUtils.project(w.targetLocation, 
                w.targetHeading + (Math.PI / 2), 
                25),
            DiaUtils.project(w.targetLocation, 
                w.targetHeading - (Math.PI / 2), 25), 
            Color.darkGray));
        _renderables.addAll(Arrays.asList(
            RoboGraphic.drawArrowHead(DiaUtils.project(
                w.targetLocation, w.effectiveHeading(),
                25), 10, w.effectiveHeading(), Color.darkGray)));
        _renderables.add(RoboGraphic.drawLine(w.targetLocation,
            DiaUtils.project(w.targetLocation, drawBearing, 
                drawDistance),
            Color.red));
        _renderables.add(RoboGraphic.drawCircleFilled(
            w.waveBreakLocation(), Color.red, 4));
        _renderables.add(RoboGraphic.drawPoint(w.targetLocation, 
            Color.red));
        _renderables.add(RoboGraphic.drawText(
            "" + (currentTime-w.fireTime), 
            w.targetLocation.x - 8, 
            w.targetLocation.y - 20, Color.white));
    }

    private void printCurrentGun() {
        System.out.println("Current gun: " + _currentGun.getLabel() + 
            " (" + DiaUtils.round(_virtualGuns.getRating(
                _currentGun, _opponentName) * 100, 2) +
            ")");
    }
}
