package voidious.gun;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import ags.utils.KdTree.Entry;

import robocode.util.Utils;

import voidious.gfx.ColoredValueSet;
import voidious.gfx.RoboGraphic;
import voidious.utils.BattleField;
import voidious.utils.KnnView;
import voidious.utils.DiaUtils;
import voidious.utils.DiaWave;
import voidious.utils.TimestampedFiringAngleSet;

public class MeleeGun {
    private static final String VIEW_NAME = "Melee";
    private static final int K_SIZE = 100;

    private Map<String, EnemyGunData> _enemies;
    private BattleField _battleField;
    private Vector<RoboGraphic> _renderables;
    private int _enemiesTotal;

    public MeleeGun(Map<String, EnemyGunData> enemies,
            BattleField battleField, Vector<RoboGraphic> renderables,
            int enemiesTotal) {
        _enemies = enemies;
        _battleField = battleField;
        _renderables = renderables;
        _enemiesTotal = enemiesTotal;
    }

    public double aimAtEveryone(Point2D.Double myNextLocation,
            long currentTime, int enemiesAlive, double bulletPower,
            EnemyGunData closestBot, boolean painting) {
        // TODO: parallel arrays is kinda ugly
        List<DiaWave> firingWaves = new ArrayList<DiaWave>();
        List<Double> firingAngles = new ArrayList<Double>();
        List<Double> firingDistances = new ArrayList<Double>();
        List<Double> bandwidths = new ArrayList<Double>();

        Iterator<EnemyGunData> enemyDataIterator = _enemies.values().iterator();
        while (enemyDataIterator.hasNext()) {
            EnemyGunData enemyData = enemyDataIterator.next();
            if (enemyData.alive && enemyData.views.get(VIEW_NAME).size() >= 10
                    && enemyData.lastWaveFired != null) {
                DiaWave w = enemyData.lastWaveFired;
                w.setBulletPower(bulletPower);
                KnnView<TimestampedFiringAngleSet> view =
                    enemyData.views.get(VIEW_NAME);
                int kSize = (int)Math.min(view.size() / 10, 
                    K_SIZE / enemiesAlive);
                List<Entry<TimestampedFiringAngleSet>> nearestNeighbors =
                    view.nearestNeighbors(w, DiamondFist.AIMING, kSize);
                if (nearestNeighbors == null) { continue; }

                int numScans = nearestNeighbors.size();
                for (int x = 0; x < numScans; x++) {
                    Point2D.Double dispVector = nearestNeighbors.get(x)
                        .value.displacementVector;
                    Point2D.Double projectedLocation = w.projectLocationBlind(
                        myNextLocation, dispVector, currentTime);
                    if (_battleField.rectangle.contains(projectedLocation)) {
                        double thisDistance = 
                            myNextLocation.distance(projectedLocation);
                        firingAngles.add(
                            DiaUtils.absoluteBearing(myNextLocation, 
                                projectedLocation));
                        firingWaves.add(w);
                        firingDistances.add(thisDistance);
                        bandwidths.add(DiaUtils.botWidthAimAngle(thisDistance));
                    }
                }                
            }
        }

        double bestAngle = DiamondFist.NO_FIRING_ANGLE;
        double bestDensity = Double.NEGATIVE_INFINITY;
        DiaWave bestWave = null;
        Double[] firingDistancesArray = firingDistances.toArray(new Double[0]);
        Double[] firingAnglesArray = firingAngles.toArray(new Double[0]);
        Double[] bandwidthsArray = bandwidths.toArray(new Double[0]);
        ColoredValueSet cvs = new ColoredValueSet();

        for (int x = 0; x < firingAnglesArray.length; x++) {
            double xFiringAngle = firingAnglesArray[x];
            double xDensity = 0;

            for (int y = 0; y < firingAnglesArray.length; y++) {
                double ux = Utils.normalRelativeAngle(
                        xFiringAngle - firingAnglesArray[y]) 
                    / bandwidthsArray[y];

                // Gaussian
                xDensity += Math.exp(-0.5 * ux * ux)
                    / firingDistancesArray[y];

                // Quartic
//                if (Math.abs(ux) <= 1) {
//                    xDensity += DiaUtils.square(1 - DiaUtils.square(ux))
//                        / firingDistancesArray[y];
//                }
            }

            if (xDensity > bestDensity) {
                bestAngle = xFiringAngle;
                bestDensity = xDensity;
                bestWave = firingWaves.get(x);
            }

            if (painting) {
                cvs.addValue(xDensity, xFiringAngle);
            }
        }

        if (firingAngles.isEmpty() ||
                bestAngle == DiamondFist.NO_FIRING_ANGLE) {
            return closestBot.lastWaveFired.absBearing;
        }

        if (painting) {
            double bandwidth = DiaUtils.botWidthAimAngle(
                myNextLocation.distance(closestBot.location));
            DiamondFist.drawGunAngles(_renderables, bestWave, cvs, bestAngle, 
                bandwidth);
        }

        return Utils.normalAbsoluteAngle(bestAngle);
    }

    public List<KnnView<TimestampedFiringAngleSet>> newDataViews() {
        List<KnnView<TimestampedFiringAngleSet>> views =
            new ArrayList<KnnView<TimestampedFiringAngleSet>>();
        MeleeFormula distanceMelee = new MeleeFormula(_enemiesTotal);
        KnnView<TimestampedFiringAngleSet> meleeView =
            new KnnView<TimestampedFiringAngleSet>(distanceMelee)
                .setK(K_SIZE)
                .visitsOn()
                .virtualWavesOn()
                .meleeOn()
                .setName(VIEW_NAME);
        views.add(meleeView);
        return views;
    }
}
