package voidious.move;

import java.awt.geom.Point2D;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.awt.Graphics2D;
import java.awt.Color;

import ags.utils.KdTree.Entry;
import robocode.*;
import robocode.util.Utils;
import voidious.Diamond;
import voidious.move.EnemyMoveData;
import voidious.utils.BattleField;
import voidious.utils.DiaUtils;
import voidious.utils.DiaWave.BulletShadow;
import voidious.utils.ErrorLogger;
import voidious.utils.Predictor;
import voidious.utils.RobotState;
import voidious.utils.DiaWave;
import voidious.utils.KnnView;
import voidious.utils.TimestampedGuessFactor;
import voidious.utils.geom.Circle;
import voidious.utils.geom.LineSeg;
import voidious.gfx.RoboGraphic;
import voidious.gfx.RoboPainter;
import voidious.gun.FireListener;

/**
 * Copyright (c) 2009-2011 - Voidious
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 *    1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software.
 *
 *    2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 *    3. This notice may not be removed or altered from any source
 *    distribution.
 */

//1v1 TODO: adjust for wall hits in enemy wave detection
//1v1 TODO: rammer detection

public class DiamondWhoosh implements RoboPainter, FireListener {
    private static final boolean ENABLE_DEBUGGING_GRAPHICS = true;
    private static final double CURRENT_DESTINATION_BIAS = 0.8;
    private static final int RECENT_LOCATIONS_TO_STORE = 50; 
    private static final double DIRECTION_CHANGE_THRESHOLD = Math.PI / 2;
    private static final long NUM_SLICES_BOT = 100;
    private static final int WAVES_TO_SURF = 2;
    private static final double FEARFUL_DISTANCING_EXPONENT = 4;
    private static final double NORMAL_DISTANCING_EXPONENT = 2;
    private static final double WALL_STICK = 160;
    private static final double DEFAULT_ATTACK_ANGLE = -1.047;
    private static final double MAX_ATTACK_ANGLE = Math.PI * .35;
    private static final double FLATTENER_HIT_THRESHOLD = 6.0;
    private static final int WAVE_MATCH_THRESHOLD = 50;
    private static final double NON_ZERO_VELOCITY_THRESHOLD = 0.1;

    private static final int CLOCKWISE_OPTION = 1;
    private static final int STOP_OPTION = 0;
    private static final int COUNTERCLOCKWISE_OPTION = -1;
    private static final double DEFAULT_ENERGY = 100;
    private static final double FAR_AWAY = 50000;
    private static final Color SHADOW_COLOR = new Color(50, 255, 50);
    private static final int WAVES_TO_DRAW = 5;

    private static final double BOT_WIDTH = 36;
    private static final double BOT_HALF_WIDTH = 18;
    private static final double MAX_WAVE_INTERCEPT_OFFSET = 
        BOT_HALF_WIDTH / Math.cos(Math.PI / 4);
    private static final double TYPICAL_DISTANCE = 465;
    private static final double TYPICAL_ESCAPE_RANGE = 0.98;
    private static final int NO_SURFABLE_WAVES = 0;
    private static final DiaWave NO_WAVE_FOUND = null;
    private static final boolean OBSERVE_WALL_HITS = false;

    private Diamond _robot;
    private Map<String, EnemyMoveData> _enemies; 
    private Destination _currentDestination;
    private double _currentHeading;
    private double _previousHeading;
    private long _timeSinceReverseDirection;
    private long _timeSinceVelocityChange;

    private LinkedList<OldLocation> _recentLocations;
    private double _myEnergy;
    private long _currentTime;
    private int _enemiesAlive;

    private double _desiredDistance;
    private double _fearDistance;
    private double _smoothAwayDistance;
    private double _wallStick = WALL_STICK;
    private int _lastMovementChoice = CLOCKWISE_OPTION;
    private List<MovementChoice> _movementOptions;
    private MovementChoice _optionCounterClockwise;
    private MovementChoice _optionStop;
    private MovementChoice _optionClockwise;
    private DistanceController _currentDistancer;
    private String _opponentName;
    private DiaWave _lastWaveSurfed;
    private double _lastVelocity;
    private double _previousVelocity;
    private int _flattenerToggleTimer;
    private double _lastNonZeroVelocity;
    private LinkedList<DiaWave> _enemyWaves;
    private LinkedList<DiaWave> _virtualWaves;
    private List<LinkedList<DiaWave>> _waveLists;
    private LinkedList<DiaWave> _potentialWaves;
    private List<DiaBullet> _firedBullets;
    private DiaWave _imaginaryWave;
    private int _imaginaryWaveIndex;
    private LinkedList<Point2D.Double> _pastLocations;
    private BattleField _battleField;

    private Vector<RoboGraphic> _renderables;
    private boolean _painting;
    private boolean _robocodePainting;
    private String _className = "DiamondWhoosh";

    public DiamondWhoosh(Diamond robot) {
        _robot = robot;
        _enemies = new HashMap<String, EnemyMoveData>();
        _battleField = new BattleField(_robot.getBattleFieldWidth(),
            _robot.getBattleFieldHeight());
        _recentLocations = new LinkedList<OldLocation>();
        _enemyWaves = new LinkedList<DiaWave>();
        _virtualWaves = new LinkedList<DiaWave>();
        _potentialWaves = new LinkedList<DiaWave>();
        _firedBullets = new ArrayList<DiaBullet>();
        _waveLists = new ArrayList<LinkedList<DiaWave>>();
        _waveLists.add(_enemyWaves);
        _waveLists.add(_virtualWaves);
        _imaginaryWave = NO_WAVE_FOUND;
        _pastLocations = new LinkedList<Point2D.Double>();
        _previousVelocity = _robot.getVelocity();

        _currentDistancer = new BasicDistancing();
        _opponentName = "";
        initializeMovementOptions();

        _renderables = new Vector<RoboGraphic>();
        _painting = false;
        _robocodePainting = false;
    }

    public void initRound(Diamond robot) {
        _robot = robot;
        for (EnemyMoveData enemyData : _enemies.values()) {
            enemyData.energy = DEFAULT_ENERGY;
            enemyData.distance = FAR_AWAY;
            enemyData.alive = true;
            enemyData.clearDistancesSq();
            enemyData.lastTimeHit = Long.MIN_VALUE;
            enemyData.lastTimeClosest = Long.MIN_VALUE;
            enemyData.pastLocations.clear();
            enemyData.raw1v1ShotsFiredThisRound = 0;
            enemyData.raw1v1ShotsHitThisRound = 0;
            enemyData.weighted1v1ShotsHitThisRound = 0;
            enemyData.clearNeighborCache();
            enemyData.lastBulletPower = 0;
        }
        _previousVelocity = _robot.getVelocity();
        _currentTime = 0;
        _currentDestination = new Destination(myLocation(),
            Double.POSITIVE_INFINITY, 0);
        _timeSinceReverseDirection = 0;
        _timeSinceVelocityChange = 0;
        _currentHeading = _previousHeading = _robot.getHeadingRadians();
        _enemiesAlive = _robot.getOthers();
        _lastNonZeroVelocity = 0;

        _recentLocations.clear();
        _enemyWaves.clear();
        _virtualWaves.clear();
        _potentialWaves.clear();
        _imaginaryWave = NO_WAVE_FOUND;
        _pastLocations.clear();
        _firedBullets.clear();
        _renderables.clear();
    }

    private Point2D.Double myLocation() {
        return new Point2D.Double(_robot.getX(), _robot.getY());
    }

    public void execute() {
        DiaUtils.log(_className, "execute", "", true);
        _myEnergy = _robot.getEnergy();
        _currentTime = _robot.getTime();
        _previousHeading = _currentHeading;
        _enemiesAlive = _robot.getOthers();
        if (Math.abs(_robot.getVelocity()) > NON_ZERO_VELOCITY_THRESHOLD) {
            _lastNonZeroVelocity = _robot.getVelocity();
        }
        _currentHeading = Utils.normalAbsoluteAngle(
            _robot.getHeadingRadians() + 
                (_lastNonZeroVelocity < 0 ? Math.PI : 0));
        for (EnemyMoveData enemyData : _enemies.values()) {
            if (enemyData.alive) {
                if (enemyData.lastScanTime < _currentTime) {
                    enemyData.pastLocations.add(enemyData.location);
                }
                enemyData.timeAliveTogether++;
                enemyData.totalDistance += enemyData.distance;
            }
        }
        updateBotDistances();
        updateDamageFactors();

        move();
        _pastLocations.addFirst(Predictor.nextLocation(_robot));
        _currentTime = _currentTime + 1;
        _previousVelocity = _robot.getVelocity();
        DiaUtils.log(_className, "execute", "", false);
    }

    private void move() {
        if (is1v1()) {
            checkActiveEnemyWaves();
            move1v1();
        } else {
            updateTimers();
            moveMelee();
        }
    }

    private void move1v1() {
        evaluateDistancing();
        evaluateFlattener();
        surf();
    }

    private void moveMelee() {
        Point2D.Double myLocation = myLocation();
        if (_currentTime % 5 == 0) {
            for (int x = 0; x < 5; x++) {
                _recentLocations.addFirst(
                    new OldLocation(DiaUtils.project(myLocation, 
                        Math.random() * Math.PI * 2, 
                        5 + Math.random() * Math.random() * 200), 
                        _currentTime));
            }
            while (_recentLocations.size() > RECENT_LOCATIONS_TO_STORE) {
                _recentLocations.removeLast();
            }
        }

        if (_enemies.size() == 0) { return; }

        ArrayList<Destination> possibleDestinations = 
            new ArrayList<Destination>();

        possibleDestinations.addAll(generatePointsAroundBot());

        if (myLocation.distance(_currentDestination.location) <=
            myLocation.distance(_enemies.get(closestBot()).location)) {
            _currentDestination.goAngle = DiaUtils.absoluteBearing(myLocation, 
                _currentDestination.location);
            _currentDestination.risk = CURRENT_DESTINATION_BIAS *
                evaluateRisk(_currentDestination.location, 
                    _currentDestination.goAngle);
            possibleDestinations.add(_currentDestination);
        }

        Destination nextDestination;
        do { 
            nextDestination = safestDestination(possibleDestinations);
            possibleDestinations.remove(nextDestination);
        } while (wouldHitWall(nextDestination));

        double goAngle = 
            DiaUtils.absoluteBearing(myLocation, nextDestination.location);
        DiaUtils.setBackAsFront(_robot, goAngle);

        _currentDestination = nextDestination;

        if (paintStatus()) {
            possibleDestinations.add(_currentDestination);
            drawRisks(possibleDestinations);
        }
    }

    public void onScannedRobot(ScannedRobotEvent e) {
        DiaUtils.log(_className, "onScannedRobot", "", true);

        Point2D.Double myLocation = myLocation();
        String botName = e.getName();
        _opponentName = botName;
        double absBearing = Utils.normalAbsoluteAngle(
            e.getBearingRadians() + _robot.getHeadingRadians());
        Point2D.Double enemyLocation = DiaUtils.project(myLocation, 
            absBearing, e.getDistance());

        EnemyMoveData enemyData;
        double previousEnergy;
        if (_enemies.containsKey(botName)) {
            enemyData = _enemies.get(botName);
            previousEnergy = enemyData.energy;
            enemyData.energy = e.getEnergy();
            enemyData.distance = e.getDistance();
            enemyData.location = enemyLocation;
            enemyData.heading = e.getHeadingRadians();
            enemyData.velocity = e.getVelocity();
            enemyData.absBearing = absBearing;
            enemyData.lastScanTime = _currentTime;
        } else {
            previousEnergy = e.getEnergy();
            enemyData = new EnemyMoveData(botName, e.getDistance(), 
                e.getEnergy(), enemyLocation, e.getHeadingRadians(), absBearing, 
                _currentTime);
            _enemies.put(botName, enemyData);
        }
        enemyData.pastLocations.add(enemyLocation);

        if (is1v1()) {
            updateTimers();

            double currentEnergy = e.getEnergy();
            boolean realShotFired = false;
            double bulletPower = previousEnergy - currentEnergy;
            if (bulletPower > 0.09 && bulletPower < 3.01) {
                realShotFired = true;
                enemyData.raw1v1ShotsFired++;
                enemyData.raw1v1ShotsFiredThisRound++;
            }
            fireEnemyWave(realShotFired, botName, bulletPower);
        }

        DiaUtils.log(_className, "onScannedRobot", "", false);
    }

    public void onRobotDeath(RobotDeathEvent e) {
        String botName = e.getName();

        try {
            _enemies.get(botName).alive = false;
        } catch (java.lang.NullPointerException npe) {
            System.out.println(
                "WARNING (move): A bot died that I never knew existed!");
        }
    }

    public void onHitByBullet(HitByBulletEvent e) {
        String botName = e.getName();

        if (_enemies.containsKey(botName)) {
            EnemyMoveData enemyData = _enemies.get(botName);
            enemyData.lastTimeHit = _robot.getTime();
            enemyData.damageTaken += 
               Rules.getBulletDamage(e.getBullet().getPower());
            enemyData.totalBulletPower += e.getBullet().getPower();
            enemyData.totalTimesHit++;
            enemyData.energy += Rules.getBulletHitBonus(e.getBullet().getPower());
            DiaWave hitWave = processBullet(e.getBullet());
            if (hitWave != NO_WAVE_FOUND) {
                double thisHit = (hitWave.targetDistance
                    / TYPICAL_DISTANCE)
                    * (hitWave.escapeAngleRange() / TYPICAL_ESCAPE_RANGE);
                enemyData.weighted1v1ShotsHit += thisHit;
                enemyData.weighted1v1ShotsHitThisRound += thisHit;
                enemyData.raw1v1ShotsHit++;
                enemyData.raw1v1ShotsHitThisRound++;
            }
        } else {
            System.out.println("WARNING (move): A bot shot me that I never " +
                "knew existed! (" + botName + ")");
        }

        if (is1v1()) {
            duelOpponent().clearNeighborCache();
        }
    }

    public void onBulletHit(BulletHitEvent e) {
        String botName = e.getName();

        try {
            EnemyMoveData enemyData = _enemies.get(botName);
            enemyData.energy -= 
                Rules.getBulletDamage(e.getBullet().getPower());
            enemyData.damageGiven += 
                Rules.getBulletDamage(e.getBullet().getPower());
        } catch (java.lang.NullPointerException npe) {
            System.out.println(
                "WARNING (move): One of my bullets hit a bot that I never " +
                "knew existed!");
        }
    }

    public void onBulletHitBullet(BulletHitBulletEvent e) {
        String botName = e.getHitBullet().getName();

        if (_enemies.containsKey(botName)) {
            EnemyMoveData enemyData = _enemies.get(botName);
            DiaWave hitWave = processBullet(e.getHitBullet());
            if (hitWave != NO_WAVE_FOUND) {
                enemyData.raw1v1ShotsFired--; 
                enemyData.raw1v1ShotsFiredThisRound--;
            }
        } else {
            System.out.println("WARNING (move): One of my bullets hit a " +
                "bullet from a bot that I never knew existed!");
        }

        if (is1v1()) {
            removeFiredBullet(e);
            for (DiaWave w : _enemyWaves) {
                w.shadows.clear();
                for (DiaBullet b : _firedBullets) {
                    setShadows(w, b);
                }
            }
            duelOpponent().clearNeighborCache();
        }
    }

    public void onPaint(Graphics2D g) {
        if (paintStatus()) {
            for (RoboGraphic renderable : _renderables) {
                renderable.render(g);
            }
            _renderables.clear();
        }
    }

    public void onWin(WinEvent e) {
        roundOver();
    }

    public void onDeath(DeathEvent e) {
        roundOver();
    }

    private void roundOver() {
        if (is1v1()) {
            duelOpponent().lastRoundNormalized1v1HitPercentage = 
                normalizedEnemyHitPercentageThisRound();
            System.out.println("Enemy normalized hit %: " +
                DiaUtils.round(normalizedEnemyHitPercentage(), 2));
            System.out.println("Enemy raw hit %: " +
                DiaUtils.round(rawEnemyHitPercentage(), 2));
            System.out.println("Flattener enabled: " +
                duelOpponent().flattenerEnabled);
        }
    }

    private boolean wouldHitWall(Destination d) {
        long ticksAhead = 5;
        double heading = _robot.getHeadingRadians();
        double velocity = _robot.getVelocity();
        RobotState state = new RobotState(myLocation(), heading, velocity);
        boolean hitWall = false;

        for (int x = 0; x < ticksAhead && !hitWall; x++) {
            state = Predictor.nextLocation(state.location, state.velocity, 8.0, 
                state.heading, DiaUtils.absoluteBearing(state.location, 
                d.location), _currentTime + x, Predictor.IGNORE_WALLS,
                _battleField);
            if (!_battleField.rectangle.contains(state.location)) {
                hitWall = true;
            }
        }

        return hitWall;
    }

    private ArrayList<Destination> generatePointsAroundBot() {
        ArrayList<Destination> destinations = new ArrayList<Destination>();
        double movementStick = Math.min(100 + Math.random() * 100,
            distanceToClosestBot());

        double sliceSize = (2 * Math.PI) / NUM_SLICES_BOT;
        Point2D.Double myLocation = myLocation();
        for (int x = 0; x < NUM_SLICES_BOT; x++) {
            double angle = x * sliceSize;
            Point2D.Double dest = DiaUtils.project(myLocation, angle, 
                movementStick);
            dest.x = DiaUtils.limit(BOT_HALF_WIDTH, dest.x, 
                _battleField.width - BOT_HALF_WIDTH);
            dest.y = DiaUtils.limit(BOT_HALF_WIDTH, dest.y, 
                _battleField.height - BOT_HALF_WIDTH);
            destinations.add(new Destination(dest, evaluateRisk(dest, angle),
                angle));
        }

        return destinations;
    }

    private double evaluateRisk(Point2D.Double destination, double goAngle) {
        double risk = 0;
        for (EnemyMoveData enemyData : _enemies.values()) {
            if (enemyData.alive) {
                double botRisk = 0;
                double distanceSq = destination.distanceSq(enemyData.location);
                botRisk = DiaUtils.limit(0.25, enemyData.energy / _myEnergy, 4)
                        * (1 + Math.abs(
                            Math.cos(enemyData.absBearing - goAngle)))
                        * enemyData.damageFactor
//                        * (20 + Math.sqrt(cornerDistance(destination)))
//                        * (_currentTime - enemyData.lastTimeHit < 100 ? 2 : 1)
//                        * (100 + enemyData.damageTaken) // could cache this
                        / (distanceSq
                            * (enemyData.botsCloser(distanceSq * .8) + 1)  
//                            * (100 + enemyData.damageGiven) // could cache this
//                            * (1 + enemyData.sumAbsBearingRisk(destination))
                            );
/*
                for (int x = 10; 
                     x < Math.min(101, enemyData.pastLocations.size()); 
                     x += 10) {
                    botRisk *= 1 + (250 / enemyData.getPastLocation(x)
                                .distanceSq(destination));
                }
*/
                risk += botRisk;
            }
        }

        double randomRisk = 0;
        for (OldLocation oldLocation : _recentLocations) {
            randomRisk += 30.0 / oldLocation.location.distanceSq(destination);
        }
        risk *= 1 + randomRisk;

        return risk;
    }

    private Destination safestDestination(
        ArrayList<Destination> possibleDestinations) {

        double lowestRisk = Double.POSITIVE_INFINITY;
        Destination safest = null;

        for (Destination destination : possibleDestinations) {
            if (destination.risk < lowestRisk) {
                lowestRisk = destination.risk;
                safest = destination;
            }
        }

        if (safest == null) {
            String error = "No safe destinations found, there must be a bug " +
                "in the risk evaluation.\n"
                + "_myLocation: (" + DiaUtils.round(myLocation().x, 1) + ", "
                    + DiaUtils.round(myLocation().y, 1) + "\n"
                + "_myEnergy: " + _myEnergy + "\n"
                + "getOthers(): " + _robot.getOthers();
            ErrorLogger.logError(error);

            safest = _currentDestination;
        }

        return safest;
    }

    private String closestBot() {
        double closestDistance = Double.POSITIVE_INFINITY;
        String closestBot = null;
        for (String botName : _enemies.keySet()) {
            EnemyMoveData enemyData = _enemies.get(botName);
            if (enemyData.alive && enemyData.distance < closestDistance) {
                closestDistance = enemyData.distance;
                closestBot = botName;
            }
        }

        return closestBot;
    }

    private double distanceToClosestBot() {
        double closestDistance = Double.POSITIVE_INFINITY;
        for (String botName : _enemies.keySet()) {
            EnemyMoveData enemyData = _enemies.get(botName);
            if (enemyData.alive && enemyData.distance < closestDistance) {
                closestDistance = enemyData.distance;
            }
        }

        return closestDistance;
    }

    private void updateBotDistances() {
        if (_enemies.size() > 1) {
            Point2D.Double myLocation = myLocation();
            String[] botNames = new String[_enemies.size()];
            _enemies.keySet().toArray(botNames);
            for (int x = 0; x < botNames.length; x++) {
                EnemyMoveData enemyData1 = _enemies.get(botNames[x]);
                for (int y = x + 1; y < botNames.length; y++) {
                    EnemyMoveData enemyData2 = _enemies.get(botNames[y]);
                    if (enemyData1.alive && enemyData2.alive) {
                        double distanceSq = 
                            enemyData1.location.distanceSq(enemyData2.location);
                        enemyData1.setBotDistanceSq(botNames[y], distanceSq);
                        enemyData2.setBotDistanceSq(botNames[x], distanceSq);
                    } else {
                        if (!enemyData1.alive) {
                            enemyData2.removeDistanceSq(botNames[x]);
                        }
                        if (!enemyData2.alive) {
                            enemyData1.removeDistanceSq(botNames[y]);
                        }
                    }
                }
                enemyData1.distance = myLocation.distance(enemyData1.location);
            }
        }
    }

    private void updateDamageFactors() {
        if (_enemies.size() > 1) {
            for (EnemyMoveData enemyData : _enemies.values()) {
                enemyData.damageFactor = 
                    ((enemyData.damageTaken + 10) / (enemyData.damageGiven + 10))
                    * enemyData.totalDistance
                    / DiaUtils.square(enemyData.timeAliveTogether);
            }
        }
    }

    private void surf() {
        if (_opponentName.equals("")) {
            return;
        }

        Point2D.Double myLocation = myLocation();
        RobotState currentState = new RobotState(myLocation,
            _robot.getHeadingRadians(), _robot.getVelocity(), _robot.getTime());
        boolean goingClockwise = (_lastMovementChoice == CLOCKWISE_OPTION);

        Collections.sort(_movementOptions);

        double bestSurfDanger = Double.POSITIVE_INFINITY;
        for (int x = 0; x < _movementOptions.size(); x++) {
            MovementChoice mc = _movementOptions.get(x);
            double thisDanger = checkDanger(currentState, 
                mc.getMovementOption(), goingClockwise, DiaWave.FIRST_WAVE, 
                WAVES_TO_SURF, bestSurfDanger);
            mc.lastDanger = thisDanger;

            bestSurfDanger = Math.min(bestSurfDanger, thisDanger);
        }

        double orbitCounterClockwiseDanger = _optionCounterClockwise.lastDanger;
        double stopDanger = _optionStop.lastDanger;
        double orbitClockwiseDanger = _optionClockwise.lastDanger;

        int goOrientation = _lastMovementChoice;

        DiaWave orbitWave = findSurfableWave(DiaWave.FIRST_WAVE);
        if (orbitWave != _lastWaveSurfed) {
            duelOpponent().clearNeighborCache();
            _lastWaveSurfed = orbitWave;
        }

        double orbitAbsBearing, distanceToClosestWaveSource;
        Point2D.Double orbitPoint;

        if (orbitWave == null) {
            orbitPoint = _enemies.get(_opponentName).location;
        } else {
            orbitPoint = orbitWave.sourceLocation;
        }
        distanceToClosestWaveSource = myLocation.distance(orbitPoint);
        orbitAbsBearing = DiaUtils.absoluteBearing(orbitPoint, myLocation);

        double goAngle;
        if (stopDanger == NO_SURFABLE_WAVES) {
            _robot.setMaxVelocity(8);

            double goAngleCcw = orbitAbsBearing +
                (COUNTERCLOCKWISE_OPTION * ((Math.PI / 2) + 
                    DEFAULT_ATTACK_ANGLE));
            goAngleCcw = wallSmoothing(myLocation, goAngleCcw, 
                COUNTERCLOCKWISE_OPTION, distanceToClosestWaveSource);

            double goAngleCw = orbitAbsBearing +
                (CLOCKWISE_OPTION * ((Math.PI / 2) + DEFAULT_ATTACK_ANGLE));
            goAngleCw = wallSmoothing(myLocation, goAngleCw, 
                CLOCKWISE_OPTION, distanceToClosestWaveSource);

            if (Math.abs(Utils.normalRelativeAngle(goAngleCw - orbitAbsBearing))
                < Math.abs(Utils.normalRelativeAngle(goAngleCcw - orbitAbsBearing))) {
                goOrientation = CLOCKWISE_OPTION;
                goAngle = goAngleCw;
            } else {
                goOrientation = COUNTERCLOCKWISE_OPTION;
                goAngle = goAngleCcw;
            }
        } else {
            _robot.setMaxVelocity(8);
            double attackAngle = 
                _currentDistancer.attackAngle(distanceToClosestWaveSource, 
                    _desiredDistance);

            if (stopDanger <= orbitCounterClockwiseDanger &&
                    stopDanger <= orbitClockwiseDanger) {
                _robot.setMaxVelocity(0);
            } else {            
                if (orbitClockwiseDanger < orbitCounterClockwiseDanger) {
                    goOrientation = CLOCKWISE_OPTION;
                } else {
                    goOrientation = COUNTERCLOCKWISE_OPTION;
                }
            }

            goAngle = orbitAbsBearing +
                (goOrientation * ((Math.PI / 2) + attackAngle));
            goAngle = wallSmoothing(myLocation, goAngle, goOrientation,
                distanceToClosestWaveSource);

            if (paintStatus()) {
                drawRawWaves();
                for (int x = 0; x < WAVES_TO_DRAW; x++) {
                    drawWaveDangers(x);
                    drawWaveShadows(x);
                }
            }
        }

        DiaUtils.setBackAsFront(_robot, goAngle);

        _lastMovementChoice = goOrientation;

    }

    private double checkDanger(RobotState startState, int movementOption, 
        boolean previouslyMovingClockwise, int surfableWaveIndex, 
        int numWavesToSurf, double cutoffDanger) {

        DiaUtils.log(_className, "checkDanger", 
            "movementOption=" + movementOption + ", " +
            "previouslyMovingClockwise=" + previouslyMovingClockwise + ", " +
            "surfableWaveIndex=" + surfableWaveIndex + ", " +
            "numWavesToSurf=" + numWavesToSurf + ", " +
            "cutoffDanger=" + cutoffDanger,
            true);

        boolean predictClockwise;
        if (movementOption == CLOCKWISE_OPTION) {
            predictClockwise = true;
        } else if (movementOption == COUNTERCLOCKWISE_OPTION) {
            predictClockwise = false;
        } else {
            predictClockwise = previouslyMovingClockwise;
        }

        DiaWave surfWave = findSurfableWave(surfableWaveIndex);

        if (surfWave == null) { 
            DiaUtils.log(_className, "checkDanger", 
                Double.toString(NO_SURFABLE_WAVES), false);

            return NO_SURFABLE_WAVES; 
        }

        double maxHitInterceptOffset =
            surfWave.bulletSpeed + MAX_WAVE_INTERCEPT_OFFSET;
        double wavePassedInterceptOffset = surfWave.bulletSpeed;
        RobotState predictedState = startState;
        RobotState passedState = startState;

        boolean wavePassed = false;
        boolean waveHit = false;

        ArrayList<RobotState> dangerStates = new ArrayList<RobotState>();

        double maxVelocity = (movementOption == STOP_OPTION) ? 0 : 8;

        do {    
            double orbitAbsBearing = DiaUtils.absoluteBearing(
                surfWave.sourceLocation, predictedState.location);
            double orbitDistance = 
                surfWave.sourceLocation.distance(predictedState.location);
            double attackAngle = _currentDistancer.attackAngle(orbitDistance, 
                _desiredDistance);
            boolean clockwiseSmoothing = predictClockwise;

            if (orbitDistance < _smoothAwayDistance) {
                clockwiseSmoothing = !clockwiseSmoothing;
            }

            predictedState = 
                Predictor.nextPerpendicularWallSmoothedLocation(
                    predictedState.location, orbitAbsBearing, 
                    predictedState.velocity, maxVelocity, 
                    predictedState.heading, attackAngle, clockwiseSmoothing, 
                    predictedState.time, _battleField, _wallStick,
                    OBSERVE_WALL_HITS);

            if (!waveHit &&
                surfWave.wavePassed(predictedState.location, 
                    predictedState.time, maxHitInterceptOffset)) {

                double waveHitInterceptOffset = surfWave.bulletSpeed + 
                    DiaUtils.preciseFrontBumperOffset(surfWave.sourceLocation, 
                        predictedState.location);

                if (surfWave.wavePassed(predictedState.location, 
                        predictedState.time, waveHitInterceptOffset)) {
                    RobotState dangerState = predictedState;
                    boolean waveGone = false;
                    double dangerMaxVelocity = 0;
                    double minGoneInterceptOffset = -BOT_HALF_WIDTH;

                    do {
                        dangerStates.add(dangerState);

                        double dangerOrbitAbsBearing = DiaUtils.absoluteBearing(
                            surfWave.sourceLocation, dangerState.location);
                        double dangerOrbitDistance = surfWave.sourceLocation
                            .distance(dangerState.location);
                        double dangerAttackAngle = 
                            _currentDistancer.attackAngle(orbitDistance, 
                                _desiredDistance);
                        boolean dangerClockwiseSmoothing = predictClockwise;

                        if (dangerOrbitDistance < _smoothAwayDistance) {
                            dangerClockwiseSmoothing =
                                !dangerClockwiseSmoothing;
                        }

                        dangerState = 
                            Predictor.nextPerpendicularWallSmoothedLocation(
                                dangerState.location, dangerOrbitAbsBearing, 
                                dangerState.velocity, dangerMaxVelocity, 
                                dangerState.heading, dangerAttackAngle, 
                                dangerClockwiseSmoothing, dangerState.time, 
                                _battleField, _wallStick, OBSERVE_WALL_HITS);

                        if (!waveGone &&
                            surfWave.wavePassed(dangerState.location,
                                dangerState.time, minGoneInterceptOffset)) {

                            if (surfWave.waveGone(dangerState.location,
                                    dangerState.time)) {
                                waveGone = true;
                            }
                        }
                    } while (!waveGone);

                    waveHit = true;
                }
            }

            if (!wavePassed && surfWave.wavePassed(predictedState.location, 
                    predictedState.time, wavePassedInterceptOffset)) {
                passedState = predictedState;
                wavePassed = true;
            }
        } while (!wavePassed);

        double[] angleAndBandwidth = surfWave.preciseBotWidth(dangerStates);
        double hitFactor = surfWave.guessFactor(angleAndBandwidth[0]);
        double bandwidth = angleAndBandwidth[1] / surfWave.maxEscapeAngle;
        double danger = 
            getDangerScore(surfWave, hitFactor, bandwidth, surfableWaveIndex);

        danger *= surfWave.shadowFactor(angleAndBandwidth[0],
            angleAndBandwidth[1]);
        danger *= Rules.getBulletDamage(surfWave.bulletPower);

        double currentDistanceToWaveSource = 
            myLocation().distance(surfWave.sourceLocation);
        double currentDistanceToWave = 
            currentDistanceToWaveSource - 
                surfWave.distanceTraveled(_robot.getTime());
        double timeToImpact = currentDistanceToWave / surfWave.bulletSpeed;

//        if (duelOpponent().flattenerEnabled) {
//            danger /= DiaUtils.square(timeToImpact);          
//        } else {
            danger /= timeToImpact;
//        }

        double firstWaveMultiplier = 1;
        if (surfableWaveIndex == DiaWave.FIRST_WAVE) {
            double predictedDistanceToWaveSource = 
                surfWave.sourceLocation.distance(passedState.location);
            double predictedDistanceToEnemy =
                _enemies.get(_opponentName).location
                    .distance(passedState.location);

            double shorterDistance = Math.min(predictedDistanceToWaveSource,
                predictedDistanceToEnemy);
            double distancingDangerBase = 
                currentDistanceToWaveSource / shorterDistance;
            double distancingDangerExponent = shorterDistance > _fearDistance ?
                NORMAL_DISTANCING_EXPONENT : FEARFUL_DISTANCING_EXPONENT;

            danger *= firstWaveMultiplier = Math.pow(distancingDangerBase, 
                distancingDangerExponent);;
        }

        if (surfableWaveIndex + 1 < numWavesToSurf && danger < cutoffDanger) {
            double nextCounterClockwiseDanger = checkDanger(passedState,
                COUNTERCLOCKWISE_OPTION, predictClockwise, 
                surfableWaveIndex + 1, numWavesToSurf, cutoffDanger);
            double nextStopDanger = checkDanger(passedState, 
                STOP_OPTION, predictClockwise, surfableWaveIndex + 1,
                numWavesToSurf, cutoffDanger);
            double nextClockwiseDanger = checkDanger(passedState,
                CLOCKWISE_OPTION, predictClockwise, 
                surfableWaveIndex + 1, numWavesToSurf, cutoffDanger);

            danger += firstWaveMultiplier * Math.min(nextCounterClockwiseDanger,
                Math.min(nextStopDanger, nextClockwiseDanger));
        }

        DiaUtils.log(_className, "checkDanger", Double.toString(danger), false);

        return danger;
    }

    private double getDangerScore(DiaWave w, Point2D.Double predictedLocation,
            int surfableWaveIndex) {
        double checkFactor = w.guessFactor(predictedLocation);
        double bandwidth = DiaUtils.botWidthAimAngle(
            predictedLocation.distance(w.sourceLocation)) / w.maxEscapeAngle;
        return getDangerScore(w, checkFactor, bandwidth, surfableWaveIndex);
    }

    private double getDangerScore(DiaWave w, double checkFactor, 
            double bandwidth, int surfableWaveIndex) {
        EnemyMoveData enemyData = _enemies.get(w.botName);

        if (enemyData.views.get(0).size() == 0) {
            return headOnDanger(checkFactor);
        }

        double totalDanger = 0;
        double totalScanWeight = 0;
        for (KnnView<TimestampedGuessFactor> view : enemyData.views) {
            if (view.enabled(normalizedEnemyHitPercentage(), 
                    duelOpponent().flattenerEnabled) && view.size() > 0) {
                List<Entry<TimestampedGuessFactor>> nearestNeighbors;
                if (view.cachedNeighbors.size() > surfableWaveIndex) {
                    nearestNeighbors =
                        view.cachedNeighbors.get(surfableWaveIndex);
                } else {
                    int thisK = DiaUtils.limit(1, view.size() / view.kDivisor,
                        view.kSize);
                    nearestNeighbors = view.nearestNeighbors(w, false, thisK);
                    view.cachedNeighbors.add(nearestNeighbors);
                }

                view.setDecayWeights(nearestNeighbors);

                double density = 0;
                double viewScanWeight = 0;
                int numScans = nearestNeighbors.size();
                for (int x = 0; x < numScans; x++) {
                    TimestampedGuessFactor tsgf = nearestNeighbors.get(x).value;
                    double xGuessFactor = tsgf.guessFactor;
                    double scanWeight = tsgf.weight
                        / Math.sqrt(nearestNeighbors.get(x).distance);

                    if (!w.shadowed(w.firingAngle(xGuessFactor))) {
                        double ux = (xGuessFactor - checkFactor) / bandwidth;
                        density += Math.exp(-0.5 * ux * ux) * scanWeight;
                    }

                    viewScanWeight += scanWeight;
                }

                totalScanWeight += viewScanWeight * view.weight;
                totalDanger += view.weight * density;
            }
        }

        return totalDanger / totalScanWeight;
    }

    private void initializeMovementOptions() {
        _movementOptions = new ArrayList<MovementChoice>();
        _movementOptions.add(_optionCounterClockwise = 
            new MovementCounterClockwise());
        _movementOptions.add(_optionStop = new MovementStop());
        _movementOptions.add(_optionClockwise = new MovementClockwise());
    }

    private void fireEnemyWave(boolean detectedEnemyWave, String botName,
            double bulletPower) {
        double currentVelocity = _robot.getVelocity();
        EnemyMoveData enemyData = _enemies.get(botName);
        int velocitySign = DiaUtils.nonZeroSign(
            Math.abs(currentVelocity) > NON_ZERO_VELOCITY_THRESHOLD ? 
                currentVelocity : _lastNonZeroVelocity);
        double accel = DiaUtils.limit(-Rules.DECELERATION, 
            DiaUtils.accel(currentVelocity, _previousVelocity), 
            Rules.ACCELERATION);
        double guessedPower = guessBulletPower(enemyData.distance, 
            enemyData.energy, _robot.getEnergy());

        Point2D.Double myLocation = myLocation();
        DiaWave futureWave = new DiaWave(botName, enemyData.location,
            myLocation, _robot.getTime() + 1, guessedPower,
            _robot.getHeadingRadians(), currentVelocity, velocitySign,
            _battleField)
                .setAccel(accel)
                .setDistance(enemyData.distance)
                .setDchangeTime(_timeSinceReverseDirection)
                .setVchangeTime(_timeSinceVelocityChange)
                .setDistanceLast8Ticks(myLocation.distance(pastLocation(8)))
                .setDistanceLast20Ticks(myLocation.distance(pastLocation(20)))
                .setDistanceLast40Ticks(myLocation.distance(pastLocation(40)))
                .setTargetEnergy(_robot.getEnergy())
                .setSourceEnergy(enemyData.energy);
        _potentialWaves.addFirst(futureWave);

        _imaginaryWaveIndex = -1;
        for (int x = 0; x < WAVES_TO_SURF; x++) {
            if (findSurfableWave(x) == NO_WAVE_FOUND) {
                _imaginaryWaveIndex = x;
                break;
            }
        }        

        double enemyGunHeat = enemyData.getGunHeat(_currentTime);
        if (_imaginaryWaveIndex >= 0 && enemyGunHeat < 0.1000001) {
            enemyData.clearNeighborCache();
            Point2D.Double aimedFromLocation = null;

            if (enemyGunHeat < 0.0000001 && _potentialWaves.size() >= 2) {
                _imaginaryWave = _potentialWaves.get(1);
                aimedFromLocation = enemyData.getPastLocation(1);
            } else {
                _imaginaryWave = futureWave;
                aimedFromLocation = enemyData.location;
                Point2D.Double sourceLocation = Predictor.nextLocation(
                    enemyData.location, enemyData.velocity, enemyData.heading);

                if (sourceLocation.distance(myLocation) < BOT_WIDTH) {
                    sourceLocation = enemyData.location;
                }

                _imaginaryWave.sourceLocation = sourceLocation;
            }

            _imaginaryWave.targetWallDistance =
                Math.min(1.5, DiaUtils.orbitalWallDistance(
                    aimedFromLocation, _imaginaryWave.targetLocation, 
                    bulletPower, _imaginaryWave.orbitDirection, _battleField));
            _imaginaryWave.targetRevWallDistance =
                Math.min(1.5, DiaUtils.orbitalWallDistance(
                    aimedFromLocation, _imaginaryWave.targetLocation, 
                    bulletPower, -_imaginaryWave.orbitDirection, _battleField));
        }

        if (_potentialWaves.size() >= 3 && _pastLocations.size() >= 3) {
            DiaWave enemyWave = _potentialWaves.get(2);
            Point2D.Double aimedFromLocation = enemyData.getPastLocation(2);
            enemyWave.sourceLocation = enemyData.getPastLocation(1);
            enemyWave.targetLocation = _pastLocations.get(2);
            enemyWave.absBearing = DiaUtils.absoluteBearing(aimedFromLocation, 
                enemyWave.targetLocation);
            enemyWave.setBulletPower(bulletPower);
            enemyWave.targetWallDistance = Math.min(1.5,
                DiaUtils.orbitalWallDistance(aimedFromLocation,
                    enemyWave.targetLocation, bulletPower,
                    enemyWave.orbitDirection, _battleField));
            enemyWave.targetRevWallDistance = Math.min(1.5,
                DiaUtils.orbitalWallDistance(aimedFromLocation,
                    enemyWave.targetLocation, bulletPower,
                    -enemyWave.orbitDirection, _battleField));

            if (detectedEnemyWave) {
                if (_imaginaryWave != NO_WAVE_FOUND) {
                    enemyData.clearNeighborCache();
                }
                _imaginaryWave = NO_WAVE_FOUND;

                enemyWave.firingWave = true;
                enemyData.lastBulletPower = bulletPower;
                enemyData.lastBulletFireTime = enemyWave.fireTime;
                for (DiaBullet b : _firedBullets) {
                    setShadows(enemyWave, b);
                }
                _enemyWaves.addLast(enemyWave);                

                double[] dataPoint = duelOpponent().bulletPowerDataPoint(
                    enemyWave.targetDistance, enemyWave.sourceEnergy,
                    enemyWave.targetEnergy);
                enemyData.powerTree.addPoint(dataPoint, bulletPower);
            } else {
                _virtualWaves.addLast(enemyWave);
            }
        }
    }

    private void checkActiveEnemyWaves() {
        int currentRound = _robot.getRoundNum();
        long currentTime = _robot.getTime();
        Point2D.Double myLocation = myLocation();

        for (LinkedList<DiaWave> waves : _waveLists) {
            Iterator<DiaWave> wavesIterator = waves.iterator();
            while (wavesIterator.hasNext()) {
                DiaWave w = wavesIterator.next();
                if (!w.processedWaveBreak &&
                        w.wavePassed(myLocation, currentTime, 0)) {
                    double visitGuessFactor = w.guessFactor(myLocation);
                    EnemyMoveData enemyData = _enemies.get(w.botName);

                    for (KnnView<TimestampedGuessFactor> view : enemyData.views) {
                        if ((w.firingWave && view.logVisits) ||
                                (!w.firingWave && view.logVirtual)) {
                            view.logWave(w, new TimestampedGuessFactor(
                                currentRound, currentTime, visitGuessFactor));
                        }
                    }

                    w.processedWaveBreak = true;
                }

                if (w.processedWaveBreak && 
                        w.waveGone(myLocation, currentTime)) {
                    wavesIterator.remove();
                }
            }
        }

        if (is1v1()) {
            Iterator<DiaBullet> bulletIterator = _firedBullets.iterator();
            while (bulletIterator.hasNext()) {
                DiaBullet firedBullet = bulletIterator.next();
                if (!_battleField.rectangle.contains(
                        firedBullet.position(_currentTime))) {
                    bulletIterator.remove();
                }
            }
        }
    }

    private DiaWave processBullet(Bullet bullet) {
        long currentTime = _robot.getTime();
        int currentRound = _robot.getRoundNum();

        Point2D.Double bulletLocation = 
            new Point2D.Double(bullet.getX(), bullet.getY());
        String botName = bullet.getName();

        DiaWave hitWave = DiaWave.findClosestWave(_enemyWaves,
            bulletLocation, _robot.getTime(), WAVE_MATCH_THRESHOLD,
            bullet.getPower());

        if (hitWave == null) {
            return NO_WAVE_FOUND; 
        }

        if (!botName.equals(hitWave.botName)) {
            return NO_WAVE_FOUND;
        }

        double hitGuessFactor = hitWave.guessFactor(bulletLocation);
        EnemyMoveData enemyData = _enemies.get(botName);
        for (KnnView<TimestampedGuessFactor> view : enemyData.views) {
            if (view.logBulletHits) {
                view.logWave(hitWave, new TimestampedGuessFactor(currentRound,
                    currentTime, hitGuessFactor));
            }
        }
        hitWave.processedBulletHit = true;

        if (paintStatus()) {
            drawBulletHit(hitWave, hitGuessFactor, bulletLocation, currentTime);
        }

        return hitWave;     
    }

    private DiaWave findSurfableWave(int waveIndex) {
        DiaUtils.log(_className, "findSurfableWave", "waveIndex=" + waveIndex,
            true);

        DiaWave surfableWave = DiaWave.findSurfableWave(_enemyWaves, waveIndex,
            myLocation(), _robot.getTime());

        if (surfableWave == null) {
            if (_imaginaryWave != NO_WAVE_FOUND &&
                    waveIndex == _imaginaryWaveIndex) {
                surfableWave = _imaginaryWave;
            } else {
                surfableWave = NO_WAVE_FOUND;
            }
        }

        DiaUtils.log(_className, "findSurfableWave", 
            (surfableWave == null ? "null" : surfableWave.toString()), false);

        return surfableWave;
    }

    private void evaluateDistancing() {
        _desiredDistance = 650;
        _fearDistance = 200;
        _smoothAwayDistance = 75;
    }

    private void evaluateFlattener() {
        EnemyMoveData enemyData = duelOpponent();
        if (enemyData == null) { return; }

        setFlattener(normalizedEnemyHitPercentage() > flattenerThreshold());
    }

    private double flattenerThreshold() {
        if (_robot.getRoundNum() == 0) {
            return 100;
        } else {
            return FLATTENER_HIT_THRESHOLD + hitPercentageMarginOfError();
        }
    }

    private double hitPercentageMarginOfError() {
        if (duelOpponent() == null) {
            return 100;
        } else {
            double hitProbability = normalizedEnemyHitPercentage() / 100;
            return 100 * 1.96 * Math.sqrt(
                hitProbability*(1-hitProbability)
                / duelOpponent().raw1v1ShotsFired);
        }
    }

    private double normalizedEnemyHitPercentage() { 
        EnemyMoveData enemyData = duelOpponent();

        return ((enemyData == null || 
                    enemyData.raw1v1ShotsFired == 0) ? 0 : 
            (((double)enemyData.weighted1v1ShotsHit / 
                enemyData.raw1v1ShotsFired) * 100.0)); 
    }

    private double normalizedEnemyHitPercentageThisRound() { 
        EnemyMoveData enemyData = duelOpponent();

        return ((enemyData == null ||
                    enemyData.raw1v1ShotsFiredThisRound == 0) ? 0 : 
            (((double)enemyData.weighted1v1ShotsHitThisRound / 
                enemyData.raw1v1ShotsFiredThisRound) * 100.0)); 
    }

    private double rawEnemyHitPercentage() { 
        EnemyMoveData enemyData = duelOpponent();

        return ((enemyData == null || 
                    enemyData.raw1v1ShotsFired == 0) ? 0 : 
            (((double)enemyData.raw1v1ShotsHit / 
                enemyData.raw1v1ShotsFired) * 100.0)); 
    }

    private void setFlattener(boolean enableFlattener) {
        boolean oldFlattenerSetting = duelOpponent().flattenerEnabled;

        int preventFlattenerThrashing = 100;
        if (_flattenerToggleTimer++ > preventFlattenerThrashing && 
            oldFlattenerSetting != enableFlattener) {
            _flattenerToggleTimer = 0;

            if (enableFlattener) { enableFlattener(); } 
            else { disableFlattener(); }

            duelOpponent().clearNeighborCache();
        }
    }

    private void enableFlattener() {
        duelOpponent().flattenerEnabled = true;

        System.out.println("Curve Flattening enabled.");
    }

    private void disableFlattener() {
        duelOpponent().flattenerEnabled = false;

        System.out.println("Curve Flattening disabled.");
    }

    private double wallSmoothing(Point2D.Double startLocation, 
        double goAngleRadians, int orientation, double currentDistance) {

        if (currentDistance < _smoothAwayDistance) {
            orientation *= -1;
        }

        double smoothedAngle = DiaUtils.wallSmoothing(_battleField,
            startLocation, goAngleRadians, orientation, _wallStick);

        return smoothedAngle;
    }

    // In FFA, fine to do this from execute().
    // In 1v1, want to do it before firing waves, so we do
    // it from onScannedRobot.
    private void updateTimers() {
        if (Math.abs(Utils.normalRelativeAngle(_currentHeading - 
                _previousHeading)) > DIRECTION_CHANGE_THRESHOLD) {
            _timeSinceReverseDirection = 0;
        } else {
            _timeSinceReverseDirection++;
        }

        double newVelocity = _robot.getVelocity();
        if (Math.abs(newVelocity - _lastVelocity) > 0.5) {
            _timeSinceVelocityChange = 0;
        } else {
            _timeSinceVelocityChange++;
        }
        _lastVelocity = newVelocity;
    }

    private void drawRisks(ArrayList<Destination> destinations) {
        double lowestRisk = Double.POSITIVE_INFINITY;
        double highestRisk = Double.NEGATIVE_INFINITY;

        double[] risks = new double[destinations.size()];
        int x = 0;
        for (Destination destination : destinations) {
            risks[x++] = destination.risk; 
            if (destination.risk < lowestRisk) {
                lowestRisk = destination.risk;
            }
            if (destination.risk > highestRisk) {
                highestRisk = destination.risk;
            }
        }

        double avg = DiaUtils.average(risks);
        double stDev = DiaUtils.standardDeviation(risks);
        for (Destination destination : destinations) {
            _renderables.add(RoboGraphic.drawCircleFilled(destination.location, 
                riskColor(destination.risk - lowestRisk, avg - lowestRisk, 
                    stDev, true, 1), 2));
        }
    }

    private void drawRawWaves() {
        long currentTime = _robot.getTime();
        for (DiaWave w : _enemyWaves) {
            _renderables.add(RoboGraphic.drawCircle(w.sourceLocation, 
                (currentTime - w.fireTime + 1) * w.bulletSpeed,
                Color.darkGray));
            _renderables.add(RoboGraphic.drawLine(w.sourceLocation, 
                DiaUtils.project(w.sourceLocation, w.absBearing, 
                    w.distanceTraveled(currentTime + 1)), Color.darkGray));
            _renderables.add(RoboGraphic.drawCircle(
                DiaUtils.project(w.sourceLocation, w.absBearing, 
                    w.distanceTraveled(currentTime + 1)),
                5, Color.darkGray));
            _renderables.add(RoboGraphic.drawCircleFilled(w.sourceLocation, 
                Color.darkGray, 4));
        }
    }

    private void drawWaveDangers(int surfableWaveIndex) {
        DiaWave surfWave = findSurfableWave(surfableWaveIndex); 
        if (surfWave == null) {
            return;
        }

        int numBins = 51;
        double halfBins = (double)(numBins - 1) / 2;

        double[] gfDangers = new double[numBins];
        boolean[] gfShadowed = new boolean[numBins];
        double minDanger = Double.POSITIVE_INFINITY;

        for (int x = 0; x <= numBins - 1; x++) {
            double gf = (x - halfBins) / halfBins;
            double bearingOffset = surfWave.orbitDirection * 
                (gf * surfWave.maxEscapeAngle);
            double absFiringAngle = surfWave.absBearing + bearingOffset;
            Point2D.Double dangerLocation = DiaUtils.project(
                surfWave.sourceLocation, absFiringAngle,
                surfWave.distanceTraveled(_currentTime));
            if (surfWave.shadowed(absFiringAngle)) {
                gfShadowed[x] = true;
                gfDangers[x] = 0;
            } else {
                gfShadowed[x] = false;
                gfDangers[x] = getDangerScore(surfWave, dangerLocation,
                    surfableWaveIndex);
            }

            if (gfDangers[x] < minDanger) {
                minDanger = gfDangers[x];
            }
        }

        double avg = DiaUtils.average(gfDangers);
        double stDev = DiaUtils.standardDeviation(gfDangers);

        for (int x = 0; x <= numBins - 1; x++) {
            double gf = (x - halfBins) / halfBins;
            double bearingOffset = surfWave.orbitDirection * 
                (gf * surfWave.maxEscapeAngle);
            Point2D.Double drawLocation = DiaUtils.project(
                surfWave.sourceLocation, surfWave.absBearing + bearingOffset,
                surfWave.distanceTraveled(_currentTime + 1) - 15);

            Color binColor = riskColor(gfDangers[x] - minDanger,
                avg - minDanger, stDev, false, 1);
            _renderables.add(RoboGraphic.drawCircleFilled(drawLocation,
                gfShadowed[x] ? SHADOW_COLOR : binColor, 2));
        }
    }

    private void drawWaveShadows(int surfableWaveIndex) {
        DiaWave surfWave = findSurfableWave(surfableWaveIndex); 
        if (surfWave == null) {
            return;
        }

        long currentTime = _robot.getTime();
        for (BulletShadow shadow : surfWave.shadows) {
            Point2D.Double p1 = DiaUtils.project(
                surfWave.sourceLocation, shadow.minAngle,
                surfWave.distanceTraveled(currentTime + 1));
            Point2D.Double p2 = DiaUtils.project(
                surfWave.sourceLocation, shadow.maxAngle,
                surfWave.distanceTraveled(currentTime + 1));
            _renderables.add(RoboGraphic.drawLine(p1, p2, SHADOW_COLOR));
        }
    }

    private void drawBulletHit(DiaWave hitWave, double hitGuessFactor,
            Point2D.Double bulletLocation, long time) {
        _renderables.add(RoboGraphic.drawLine(
            hitWave.sourceLocation,
            DiaUtils.project(hitWave.sourceLocation, hitWave.absBearing, 
                hitWave.distanceTraveled(time)), 
            Color.yellow));
        if (Math.abs(hitGuessFactor) > 0.01) {
            _renderables.add(RoboGraphic.drawLine(
                hitWave.sourceLocation, bulletLocation, 
                (hitGuessFactor > 0 ? Color.blue : Color.red)));
        }
    }

    private static Color riskColor(double risk, double avg,
        double stDev, boolean safestYellow, double maxStDev) {

        if (risk < .0000001 && safestYellow) {
            return Color.yellow;
        }

        return new Color(
            (int)DiaUtils.limit(0, 255 * (risk - (avg - maxStDev*stDev)) 
                    / (2*maxStDev*stDev), 255), 
                0,
            (int)DiaUtils.limit(0, 255 * ((avg + maxStDev*stDev) - risk) 
                    / (2*maxStDev*stDev), 255));
    }

    private boolean is1v1() {
        return (_enemiesAlive <= 1);
    }

    private EnemyMoveData duelOpponent() {
        return _enemies.get(_opponentName);
    }

    public void paintOn() {
        _painting = ENABLE_DEBUGGING_GRAPHICS;
    }

    public void paintOff() {
        _renderables.clear();
        _painting = false;
    }

    public void robocodePaintOn() {
        _robocodePainting = true;
    }

    public void robocodePaintOff() {
        _renderables.clear();
        _robocodePainting = false;
    }

    public String paintLabel() {
        return "Movement";
    }

    public boolean paintStatus() {
        return (_painting && _robocodePainting);
    }

    private static double headOnDanger(double checkFactor) {
        return 1 - DiaUtils.square(checkFactor);        
    }

    private double guessBulletPower(double distance, double enemyEnergy, 
            double myEnergy) {
        int numBullets = duelOpponent().powerTree.size();
        if (numBullets == 0) {
            return 1.9;
        } else {
            double[] searchPoint = duelOpponent()
                .bulletPowerDataPoint(distance, enemyEnergy, myEnergy);
            List<Entry<Double>> bulletPowers = duelOpponent().powerTree
                .nearestNeighbor(searchPoint, 
                    (int)Math.min(20, Math.ceil(numBullets / 3.0)), false);

            double powerTotal = 0;
            for (Entry<Double> entry : bulletPowers) {
                powerTotal += entry.value;
            }

            return DiaUtils.round(powerTotal / bulletPowers.size(), 6);
        }
    }

    private Point2D.Double pastLocation(int x) {
        return _pastLocations.get(Math.min(x, _pastLocations.size()-1));
    }

    @Override
    public void bulletFired(DiaBullet bullet) {
        if (is1v1()) {
            _firedBullets.add(bullet);
            for (DiaWave w : _enemyWaves) {
                setShadows(w, bullet);
            }
        }
    }

    // TODO: I'm fairly certain the shadow created by a bullet going from
    //       inside to outside a wave is irrelevant, so I'm ignoring it.
    //       Would be nice to prove it or fix this code.
    private void setShadows(DiaWave w, DiaBullet b) {
        if (!w.processedBulletHit &&
                w.sourceLocation.distanceSq(b.position(_currentTime)) >
                DiaUtils.square(w.distanceTraveled(_currentTime))) {
            long time = _currentTime;
            do {
                time++;
                if (w.sourceLocation.distanceSq(b.position(time)) <
                        DiaUtils.square(w.distanceTraveled(time)) &&
                        time < b.deathTime) {
                    addBulletShadows(w, b, time);
                    break;
                }
            } while (_battleField.rectangle.contains(b.position(time)));
        }
    }

    private void addBulletShadows(DiaWave w, DiaBullet b, long time) {
        Circle waveCircle1 = new Circle(w.sourceLocation,
            w.distanceTraveled(time - 1));
        Circle waveCircle2 = new Circle(w.sourceLocation,
            w.distanceTraveled(time));
        Point2D.Double bulletPoint1 = b.position(time-1);
        Point2D.Double bulletPoint2 = b.position(time);
        LineSeg bulletSeg = new LineSeg(bulletPoint1, bulletPoint2);
        Point2D.Double[] xPoints1 = waveCircle1.intersects(bulletSeg);
        Point2D.Double[] xPoints2 = waveCircle2.intersects(bulletSeg);

        if (xPoints1[0] == null && xPoints2[0] == null) {
            // whole bullet line segment is between two wave circles,
            // shadow cast along all firing angles covered by line segment
            w.castShadow(bulletPoint1, bulletPoint2);
            // TODO: It's possible the previous bullet line intersected the
            //       inner wave circle (then the outer wave circle) last tick
            //       and cast 2 shadows, while remaining outside the wave on
            //       either endpoint. Should be super rare and probably could
            //       never be a reachable shadow anyway.
        } else if (xPoints1[0] != null) {
            if (xPoints2[0] != null) {
                // line segment intersects both wave circles, shadow cast
                // from one intersection to the other
                w.castShadow(xPoints2[0], xPoints1[0]);
            } else if (xPoints1[1] != null) {
                // line segment begins and ends between wave circles,
                // intersects inner circle in two spots, casts 2 shadows
                Point2D.Double intersect1, intersect2;
                if (bulletPoint1.distanceSq(xPoints1[0]) <
                    bulletPoint1.distanceSq(xPoints1[1])) {
                    intersect1 = xPoints1[0];
                    intersect2 = xPoints1[1];
                } else {
                    intersect1 = xPoints1[1];
                    intersect2 = xPoints1[0];
                }
                w.castShadow(bulletPoint1, intersect1);
                w.castShadow(intersect2, bulletPoint2);
            } else {
                // line segment crosses inner wave circle in one spot
                w.castShadow(bulletPoint1, xPoints1[0]);
            }
        } else if (xPoints2[0] != null) {
            // line segment crosses outer wave circle in one spot
            w.castShadow(xPoints2[0], bulletPoint2);
        }
    }

    private void removeFiredBullet(BulletHitBulletEvent e) {
        double closestDistanceSq = Double.POSITIVE_INFINITY;
        DiaBullet closestFiredBullet = null;
        Point2D.Double bulletPoint = new Point2D.Double(e.getBullet().getX(),
            e.getBullet().getY());
        for (DiaBullet firedBullet : _firedBullets) {
            double thisDistanceSq =
                firedBullet.position(e.getTime()).distanceSq(bulletPoint);
            if (thisDistanceSq < closestDistanceSq) {
                closestDistanceSq = thisDistanceSq;
                closestFiredBullet = firedBullet;
            }
        }
        int bulletDistanceThreshold = 40;
        if (closestFiredBullet != null &&
                closestDistanceSq < DiaUtils.square(bulletDistanceThreshold)) {
            closestFiredBullet.deathTime = e.getTime();
        }
    }

    interface DistanceController {
        public double attackAngle(double currentDistance, double desiredDistance);
    }

    static class BasicDistancing implements DistanceController {
        public double attackAngle(double currentDistance, double desiredDistance) {
            double distanceOffset = currentDistance - desiredDistance;
            double attackAngle = DiaUtils.limit(-DiamondWhoosh.MAX_ATTACK_ANGLE, 
                (distanceOffset / desiredDistance) * 1.25,
                DiamondWhoosh.MAX_ATTACK_ANGLE);

            return attackAngle;
       }
    }

    static abstract class MovementChoice implements java.lang.Comparable<MovementChoice> {
        public double lastDanger = Double.POSITIVE_INFINITY;

        abstract public int getMovementOption();

        public int compareTo(MovementChoice m2) {
            if (lastDanger < m2.lastDanger) { return -1; }
            else if (lastDanger > m2.lastDanger) { return 1; }
            else { return 0; }
        }
    }

    static class MovementCounterClockwise extends MovementChoice {
        public int getMovementOption() {
            return DiamondWhoosh.COUNTERCLOCKWISE_OPTION;
        }
    }

    static class MovementStop extends MovementChoice {
        public int getMovementOption() {
            return DiamondWhoosh.STOP_OPTION;
        }
    }

    static class MovementClockwise extends MovementChoice {
        public int getMovementOption() {
            return DiamondWhoosh.CLOCKWISE_OPTION;
        }
    }
}
