package voidious.utils;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;

import robocode.AdvancedRobot;
import robocode.RobocodeFileWriter;

/**
 * Copyright (c) 2011 - Voidious
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 *    1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software.
 *
 *    2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 *    3. This notice may not be removed or altered from any source
 *    distribution.
 */

public class ErrorLogger {
    private static final String ERROR_LOG = "error.log";
    private static final int MAX_ERROR_SIZE = 5000;

    public static boolean enabled = false;
    private static File _outFile = null;
    private static boolean _quotaAvailable = false;
    
    public static void init(AdvancedRobot robot) {
        _outFile = robot.getDataFile(ERROR_LOG);
        _quotaAvailable = robot.getDataQuotaAvailable() > MAX_ERROR_SIZE;
    }

    public static void logException(Exception e) {
        logException(e, null);
    }

    public static void logException(Exception e, String moreInfo) {
        try {
            RobocodeFileWriter rfw;
            PrintWriter pw;
            
            if (_outFile.exists()) {
                rfw =
                    new RobocodeFileWriter(_outFile.getAbsolutePath(), true);
            } else {
                rfw = new RobocodeFileWriter(_outFile);
            }
            pw = new PrintWriter(rfw);
            e.printStackTrace(pw);
            if (moreInfo != null && !moreInfo.equals("")) {
                pw.append("\n");
                pw.append(moreInfo);
                pw.append("\n");
            }
            pw.append("----");
            pw.append("\n");
            pw.close();
            rfw.close();
        } catch (IOException ioe) {
            ioe.printStackTrace();
        }
    }

    public static void logError(String output) {
        if (enabled && _quotaAvailable && output.length() < MAX_ERROR_SIZE) {
            try {
                System.out.println("ERROR:");
                System.out.println("  " + output.replaceAll("[\r\n]+", "\n  "));

                RobocodeFileWriter rfw;
                if (_outFile.exists()) {
                    rfw = new RobocodeFileWriter(_outFile.getAbsolutePath(),
                        true);
                } else {
                    rfw = new RobocodeFileWriter(_outFile);
                }
                rfw.append(output);
                rfw.append("\n\n");
                rfw.close();
            } catch (IOException ioe) {
                ioe.printStackTrace();
            }
        }
    }
}
