package voidious.utils;

import java.awt.geom.Point2D;

import robocode.util.Utils;
import voidious.Diamond;

/**
 * Copyright (c) 2009-2011 - Voidious
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 *    1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software.
 *
 *    2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 *    3. This notice may not be removed or altered from any source
 *    distribution.
 */

public class Predictor {
    public static final boolean IGNORE_WALLS = true;
    public static final boolean OBSERVE_WALL_HITS = false;
    private static final double HALF_PI = Math.PI / 2;

    public static MovSim moveSimulator;

    public static Point2D.Double nextLocation(Diamond robot) {
        if (moveSimulator == null) {
            moveSimulator = new MovSim();
        }
        MovSimStat[] next = moveSimulator.futurePos(1, robot,
            robot.getMaxVelocity());
        return new Point2D.Double(next[0].x, next[0].y);
    }

    public static Point2D.Double nextLocation(Point2D.Double botLocation,
            double velocity, double heading) {
        return new Point2D.Double(
            botLocation.x + (Math.sin(heading)*velocity),
            botLocation.y + (Math.cos(heading)*velocity));
    }

    public static RobotState nextPerpendicularLocation(
            Point2D.Double botLocation, double absBearing, double velocity,
            double heading, boolean clockwise, long currentTime,
            boolean ignoreWallHits) {
        return nextPerpendicularWallSmoothedLocation(botLocation,
            absBearing, velocity, 8.0, heading, 0,
            clockwise, currentTime, null, 0, ignoreWallHits);
    }

    public static RobotState nextPerpendicularWallSmoothedLocation(
            Point2D.Double botLocation, double absBearing, double velocity,
            double maxVelocity, double heading, double attackAngle,
            boolean clockwise, long currentTime, BattleField battleField,
            double wallStick, boolean ignoreWallHits) {
        int orientation;
        if (clockwise) { orientation = 1; }
        else { orientation = -1; }

        double goAngle = Utils.normalRelativeAngle(
            absBearing + (orientation * ((HALF_PI) + attackAngle)));
        if (wallStick != 0 && battleField != null) {
            goAngle = DiaUtils.wallSmoothing(battleField,
                botLocation, goAngle, orientation, wallStick);
        }

        return nextLocation(botLocation, velocity, maxVelocity,
            heading, goAngle, currentTime, ignoreWallHits, battleField);
    }

    public static RobotState nextLocation(Point2D.Double botLocation,
            double velocity, double maxVelocity, double heading, double goAngle,
            long currentTime, boolean ignoreWallHits, BattleField battleField) {
        MovSim movSim = getMovSim();
        double futureTurn = Utils.normalRelativeAngle(goAngle - heading);
        double futureDistance;

        if (Math.abs(futureTurn) > (HALF_PI)) {
            if (futureTurn < 0) { futureTurn = Math.PI + futureTurn; }
            else { futureTurn = -1 * (Math.PI - futureTurn); }
            futureDistance = -1000;
        } else {
            futureDistance = 1000;
        }

        int extraWallSize = 0;
        if (ignoreWallHits) {
            extraWallSize = 50000;
        }

        MovSimStat[] futureMoves = movSim.futurePos(1,
            extraWallSize + botLocation.x, extraWallSize + botLocation.y,
            velocity, maxVelocity, heading, futureDistance,
            futureTurn,
            extraWallSize * 2 + (battleField == null ? 0 : battleField.width), 
            extraWallSize * 2 + (battleField == null ? 0 : battleField.height));

        return new RobotState(new Point2D.Double(
            futureMoves[0].x - extraWallSize, futureMoves[0].y - extraWallSize),
            futureMoves[0].h, futureMoves[0].v, currentTime + 1);
    }

    public static MovSim getMovSim() {
        if (moveSimulator == null) {
            moveSimulator = new MovSim();
        }
        return moveSimulator;
    }
}
