package voidious.utils;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.Line2D;
import java.util.List;

/**
 * Copyright (c) 2009-2011 - Voidious
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 *    1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software.
 *
 *    2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 *    3. This notice may not be removed or altered from any source
 *    distribution.
 */

public class RobotState {
    public final Point2D.Double location;
    public final double heading;
    public final double velocity;
    public final long time;
    private Rectangle2D.Double _botRect;
    private List<Line2D.Double> _botSides;

    public RobotState(Point2D.Double location, double heading,
            double velocity) {
        this(location, heading, velocity, 0);
    }

    public RobotState(Point2D.Double location, double heading,
            double velocity, long time) {
        this.location = location;
        this.heading = heading;
        this.velocity = velocity;
        this.time = time;

        _botRect = null;
        _botSides = null;
    }

    public Rectangle2D.Double botRect() {
        if (_botRect == null) {
            _botRect = DiaUtils.botRect(this.location);
        }
        return _botRect;
    }

    public List<Line2D.Double> botSides() {
        if (_botSides == null) {
            _botSides = DiaUtils.botSides(this.location);
        }
        return _botSides;
    }

    public Object clone() {
        return new RobotState((Point2D.Double)location.clone(), heading, 
            velocity, time);
    }
}
