package xander.cat.drive;

import robocode.Bullet;
import robocode.HitByBulletEvent;
import xander.core.Resources;
import xander.core.event.OpponentWaveListener;
import xander.core.math.LinearEquation;
import xander.core.track.GunStats;
import xander.core.track.Wave;
import xander.gfws.data.FactorArrayParms;

public class AutoVaryWeightByAge implements OpponentWaveListener {

	private FactorArrayParms parms;
	private double noRollOppHitRatio;
	private LinearEquation rollEquation;
	private LinearEquation minWeightEquation;
	private GunStats gunStats;
	
	public AutoVaryWeightByAge(FactorArrayParms parms, 
			double noRollOppHitRatio, 
			double lowOppHitRatio, double lowRollRate, 
			double highOppHitRatio, double highRollRate) {
		this.parms = parms;
		this.noRollOppHitRatio = noRollOppHitRatio;
		this.rollEquation = new LinearEquation(lowOppHitRatio, lowRollRate, highOppHitRatio, highRollRate, lowRollRate, highRollRate);
		this.gunStats = Resources.getGunStats();
		Resources.getWaveHistory().addOpponentWaveListener(this);
	}

	public void setVaryMinWeight(double lowHRDiff, double lowHRDiffMinWeight, double highHRDiff, double highHRDiffMinWeight) {
		double minY = Math.min(lowHRDiffMinWeight, highHRDiffMinWeight);
		double maxY = Math.max(lowHRDiffMinWeight, highHRDiffMinWeight);
		this.minWeightEquation = new LinearEquation(lowHRDiff, lowHRDiffMinWeight, highHRDiff, highHRDiffMinWeight, minY, maxY);
	}
	
	@Override
	public void oppWaveCreated(Wave wave) {
		// no action required
	}

	@Override
	public void oppWaveHitBullet(Wave wave, Bullet oppBullet) {
		// no action required
	}

	@Override
	public void oppWaveHit(Wave wave) {
		double oppHitRatio = gunStats.getOverallOpponentHitRatio();
		if (oppHitRatio <= noRollOppHitRatio) {
			parms.setVaryWeightByAge(0);
		} else {
			parms.setVaryWeightByAge(rollEquation.getY(oppHitRatio));
			if (minWeightEquation != null) {
				double myHitRatio = gunStats.getOverallHitRatio();
				double hrDiff = myHitRatio - oppHitRatio;
				parms.setMinAgeWeightPercent(minWeightEquation.getY(hrDiff));
			}
		}
	}

	@Override
	public void oppNextWaveToHit(Wave wave) {
		// no action required
	}

	@Override
	public void oppBulletHit(Wave wave, HitByBulletEvent hitByBulletEvent) {
		// no action required
	}

	@Override
	public void oppWavePassing(Wave wave) {
		// no action required
	}

	@Override
	public void oppWavePassed(Wave wave) {
		// no action required
	}

	@Override
	public void oppWaveUpdated(Wave wave) {
		// no action required
	}

	@Override
	public void oppWaveDestroyed(Wave wave) {
		// no action required
	}
}
