package xander.cat.group.mirror;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

import xander.core.Resources;
import xander.core.RobotProxy;
import xander.core.drive.Direction;
import xander.core.drive.DistancingEquation;
import xander.core.drive.DriveState;
import xander.core.drive.OrbitalDrivePredictor;
import xander.core.gun.targeter.Targeter;
import xander.core.math.RCMath;
import xander.core.track.Snapshot;
import xander.core.track.Wave;

public class AntiMirrorTargeter implements Targeter {

	private MirrorPlan mirrorPlan;
	private RobotProxy robotProxy;
	private OrbitalDrivePredictor orbitalDriver;
	
	public AntiMirrorTargeter(MirrorPlan mirrorPlan, OrbitalDrivePredictor orbitalDriver) {
		this.orbitalDriver = orbitalDriver;
		this.mirrorPlan = mirrorPlan;
		this.robotProxy = Resources.getRobotProxy();
	}
	
	@Override
	public String getTargetingType() {
		return "Mirror";
	}
	
	@Override
	public boolean canAimAt(Snapshot target) {
		return true;
	}

	@Override
	public double getAim(Snapshot target,
			Snapshot myself, Wave wave) {
		// calculate our position when bullet would reach other side
		double bulletVelocity = wave.getBulletVelocity();
		double distanceToOpponent = Double.MAX_VALUE;
		double bulletTravelDistance = 0;
		DriveState myDriveState = new DriveState(robotProxy);
		long futureTime = robotProxy.getTime();
		Rectangle2D.Double battlefieldBounds = robotProxy.getBattleFieldSize();
		while (bulletTravelDistance < (distanceToOpponent - bulletVelocity*4)) { // they will be at a couple of tick behind
			bulletTravelDistance += bulletVelocity;
			Point2D.Double center = AntiMirrorDrive.getOrbitCenter(
					robotProxy.getX(), 
					robotProxy.getY(), 
					battlefieldBounds);
			Direction direction = mirrorPlan.getDirection(futureTime);
			orbitalDriver.advanceOrbitalDriveState(
					myDriveState,  
					center, 100, direction, 1, 
					AntiMirrorDrive.DRIVE_SPEED, DistancingEquation.NO_ADJUST);
			futureTime++;
			distanceToOpponent = 2 * RCMath.getDistanceBetweenPoints(
					myDriveState.getPosition().x, 
					myDriveState.getPosition().y, 
					battlefieldBounds.getCenterX(), 
					battlefieldBounds.getCenterY());
		}
		// aim gun at to hit opposing position
		double futureOppX = battlefieldBounds.getMaxX() - myDriveState.getPosition().x;
		double futureOppY = battlefieldBounds.getMaxY() - myDriveState.getPosition().y;
		double targetAngle = RCMath.getRobocodeAngle(myself.getX(), myself.getY(), futureOppX, futureOppY);
		return targetAngle;
	}
}
