package xander.core;

import java.awt.geom.Path2D;

import xander.core.gun.power.FixedPowerSelector;
import xander.core.gun.power.PowerSelector;
import xander.core.math.RCPhysics;
import xander.gfws.BalancedFactorIndexer;
import xander.gfws.FactorIndexer;

public class Configuration {

	private int snapshotHistorySize                      = 80;
	private boolean logEnemyBulletFiredStats             = false;
	private boolean autoFireOnDisabledOpponents          = true;
	private int cpuUtilizationDataPoints                 = 200;
	private boolean logComponentRunTimes                 = false;
	private boolean logDriveTimes                        = false;
	private int rollingVirtualHitRatioDepth              = 40;
	private PowerSelector disabledOpponentPowerSelector  = new FixedPowerSelector(RCPhysics.MIN_FIRE_POWER);
	private boolean clearBattleStatsOnNewVersion         = true;
	private boolean processNonFiringWaves                = false;
	private PowerSelector nonFiringPowerSelector         = null;
	private boolean usePreciseMEAForMyWaves              = false;
	private boolean usePreciseMEAForOpponentWaves        = false;
	private Path2D.Double myPreciseMEADriveBounds        = null;
	private Path2D.Double opponentPreciseMEADriveBounds  = null;
	private FactorIndexer factorIndexer                  = new BalancedFactorIndexer();

	public int getSnapshotHistorySize() {
		return snapshotHistorySize;
	}

	public void setSnapshotHistorySize(int snapshotHistorySize) {
		this.snapshotHistorySize = snapshotHistorySize;
	}

	public boolean isLogEnemyBulletFiredStats() {
		return logEnemyBulletFiredStats;
	}

	public void setLogEnemyBulletFiredStats(boolean logEnemyBulletFiredStats) {
		this.logEnemyBulletFiredStats = logEnemyBulletFiredStats;
	}

	public boolean isAutoFireOnDisabledOpponents() {
		return autoFireOnDisabledOpponents;
	}

	public void setAutoFireOnDisabledOpponents(boolean autoFireOnDisabledOpponents) {
		this.autoFireOnDisabledOpponents = autoFireOnDisabledOpponents;
	}

	public boolean isUsePreciseMEAForMyWaves() {
		return usePreciseMEAForMyWaves;
	}

	public Path2D.Double getMyPreciseMEADriveBounds() {
		return myPreciseMEADriveBounds;
	}
	
	/**
	 * Sets whether or not to store precise MEA for self with the waves in
	 * the wave history.  Setting this to true causes additional processing,
	 * so only set this to true if you are going to use it.
	 * 
	 * @param usePreciseMEAForMyWaves
	 * @param preciseMEADriveBounds
	 */
	public void setUsePreciseMEAForMyWaves(boolean usePreciseMEAForMyWaves, Path2D.Double myPreciseMEADriveBounds) {
		this.usePreciseMEAForMyWaves = usePreciseMEAForMyWaves;
		this.myPreciseMEADriveBounds = myPreciseMEADriveBounds;
	}
	
	public boolean isUsePreciseMEAForOpponentWaves() {
		return usePreciseMEAForOpponentWaves;
	}

	public Path2D.Double getOpponentPreciseMEADriveBounds() {
		return opponentPreciseMEADriveBounds;
	}
	
	/**
	 * Sets whether or not to store precise MEA of opponent with the waves in
	 * the wave history.  Setting this to true causes additional processing,
	 * so only set this to true if you are going to use it.
	 * 
	 * @param usePreciseMEAForOpponentWaves
	 * @param opponentPreciseMEADriveBounds
	 */
	public void setUsePreciseMEAForOpponentWaves(boolean usePreciseMEAForOpponentWaves, Path2D.Double opponentPreciseMEADriveBounds) {
		this.usePreciseMEAForOpponentWaves = usePreciseMEAForOpponentWaves;
		this.opponentPreciseMEADriveBounds = opponentPreciseMEADriveBounds;
	}
	
	public PowerSelector getDisabledOpponentPowerSelector() {
		return disabledOpponentPowerSelector;
	}

	public void setDisabledOpponentPowerSelector(
			PowerSelector disabledOpponentPowerSelector) {
		this.disabledOpponentPowerSelector = disabledOpponentPowerSelector;
	}

	public int getRollingVirtualHitRatioDepth() {
		return rollingVirtualHitRatioDepth;
	}

	public void setRollingVirtualHitRatioDepth(int rollingVirtualHitRatioDepth) {
		this.rollingVirtualHitRatioDepth = rollingVirtualHitRatioDepth;
	}

	public boolean isLogComponentRunTimes() {
		return logComponentRunTimes;
	}

	public void setLogComponentRunTimes(boolean logComponentRunTimes) {
		this.logComponentRunTimes = logComponentRunTimes;
	}

	public boolean isLogDriveTimes() {
		return logDriveTimes;
	}

	public void setLogDriveTimes(boolean logDriveTimes) {
		this.logDriveTimes = logDriveTimes;
	}

	public int getCPUUtilizationDataPoints() {
		return cpuUtilizationDataPoints;
	}

	public void setCPUUtilizationDataPoints(int cpuUtilizationDataPoints) {
		this.cpuUtilizationDataPoints = cpuUtilizationDataPoints;
	}

	public boolean isClearBattleStatsOnNewVersion() {
		return clearBattleStatsOnNewVersion;
	}

	public void setClearBattleStatsOnNewVersion(boolean clearBattleStatsOnNewVersion) {
		this.clearBattleStatsOnNewVersion = clearBattleStatsOnNewVersion;
	}

	public boolean isProcessNonFiringWaves() {
		return processNonFiringWaves;
	}

	public void setProcessNonFiringWaves(boolean processNonFiringWaves) {
		this.processNonFiringWaves = processNonFiringWaves;
	}

	public PowerSelector getNonFiringPowerSelector() {
		return nonFiringPowerSelector;
	}

	public void setNonFiringPowerSelector(PowerSelector nonFiringPowerSelector) {
		this.nonFiringPowerSelector = nonFiringPowerSelector;
	}

	public FactorIndexer getFactorIndexer() {
		return factorIndexer;
	}

	public void setFactorIndexer(FactorIndexer factorIndexer) {
		this.factorIndexer = factorIndexer;
	}
}
