package xander.core.gun;

import java.util.HashMap;
import java.util.Map;

import xander.core.Resources;
import xander.core.track.GunStats;
import xander.core.track.Snapshot;

public class VirtualHitRatioGunSelector implements GunSelector {

	private GunStats gunStats;
	private Map<Gun, Double> biasMap = new HashMap<Gun, Double>();
	private double rollingRatioWeight = 0.6;
	
	public VirtualHitRatioGunSelector() {
		gunStats = Resources.getGunStats();
	}
	
	public void setRollingRatioWeight(double rollingRatioWeight) {
		this.rollingRatioWeight = rollingRatioWeight;
	}
	
	public void addBias(Gun gun, double bias) {
		biasMap.put(gun, Double.valueOf(bias));
	}	
	
	@Override
	public int selectGun(Gun[] guns, Snapshot target) {
		int sGunIndex = 0;
		double sRatio = Double.NEGATIVE_INFINITY;
		//StringBuilder sb = new StringBuilder();
		for (int i=0; i<guns.length; i++) {
			Gun gun = guns[i];
			double ratio = gunStats.getVirtualHitRatio(gun.getName());
			double rollingRatio = gunStats.getRollingVirtualHitRatio(gun.getName());
			ratio = ratio * (1-rollingRatioWeight) + rollingRatio * rollingRatioWeight;
			Double bias = biasMap.get(gun);
			if (bias != null) {
				ratio += bias.doubleValue();
			}
			//sb.append("[" + gun.getName() + ":" + gunController.getVirtualBulletsFired(gun.getName()) + ":" + Logger.format(100*ratio) + "%]");
			if (ratio > sRatio && guns[i].canFireAt(target)) {
				sGunIndex = i;
				sRatio = ratio;
			}
		}
		//System.out.println(sb.toString());
		return sGunIndex;
	}
}
