package xander.gfws;

import xander.core.drive.Direction;
import xander.core.math.RCPhysics;

public class BalancedFactorIndexer implements FactorIndexer {

	@Override
	public double getPreciseFactorIndex(double factorAngle, int factors,
			Direction surfDirection, AngleRange mea) {
		if (surfDirection.isClockwise()) {
			double m = factors/(mea.getClockwiseAngle() - mea.getCounterClockwiseAngle());
			double b = -m * mea.getCounterClockwiseAngle();
			return m * factorAngle + b;
		} else {
			double m = factors/(mea.getCounterClockwiseAngle() - mea.getClockwiseAngle());
			double b = -m * mea.getClockwiseAngle();
			return m * factorAngle + b;
		}
	}

	@Override
	public int getFactorIndex(double factorAngle, int factors,
			Direction surfDirection, AngleRange mea) {
		return (int) Math.round(Math.floor(getPreciseFactorIndex(factorAngle, factors, surfDirection, mea)));
	}

	@Override
	public double getFactorAngle(double preciseFactorIndex, int factors,
			Direction surfDirection, AngleRange mea) {
		if (surfDirection.isClockwise()) {
			double m = (mea.getClockwiseAngle() - mea.getCounterClockwiseAngle())/factors;
			return m * preciseFactorIndex + mea.getCounterClockwiseAngle();
		} else {
			double m = (mea.getCounterClockwiseAngle() - mea.getClockwiseAngle())/factors;
			return m * preciseFactorIndex + mea.getClockwiseAngle();
		}
	}

	@Override
	public IndexRange getRobotIndexWidth(int factors, double distance,
			double currentRobotFactorAngle, Direction surfDirection, AngleRange mea) {
		double circ = 2*Math.PI*distance;
		double halfBodyWidthDegrees = 180d*RCPhysics.ROBOT_WIDTH/circ;
		double ccwIdx = getPreciseFactorIndex(currentRobotFactorAngle-halfBodyWidthDegrees, factors, surfDirection, mea);
		double cwIdx = getPreciseFactorIndex(currentRobotFactorAngle+halfBodyWidthDegrees, factors, surfDirection, mea);
		if (ccwIdx < cwIdx) {
			return new IndexRange(ccwIdx, cwIdx);
		} else {
			return new IndexRange(cwIdx, ccwIdx);
		}
	}

}
