package xander.gfws;

import xander.core.log.Logger;

/**
 * Stores the max escape angle range from a counter-clockwise factor angle
 * to a clockwise factor angle.  
 * 
 * @author Scott Arnold
 */
public class MEA {

	private double counterClockwiseMEA;
	private double clockwiseMEA;
	
	/**
	 * Construct a new Max Escape Angle (MEA).  Counter-clockwise
	 * MEA should be less than clockwise MEA.  Under normal circumstances, 
	 * counter-clockwise MEA should be negative, while clockwise MEA is positive, 
	 * though there can be an exception when using precise MEA and the time to 
	 * bullet hit is very low.
	 * 
	 * @param counterClockwiseMEA   max counter-clockwise factor angle.
	 * @param clockwiseMEA          max clockwise factor angle.
	 */
	public MEA(double counterClockwiseMEA, double clockwiseMEA) {
		this.counterClockwiseMEA = counterClockwiseMEA;
		this.clockwiseMEA = clockwiseMEA;
	}

	public double getCounterClockwiseMEA() {
		return counterClockwiseMEA;
	}

	public double getClockwiseMEA() {
		return clockwiseMEA;
	}
	
	public double getArcSize() {
		return clockwiseMEA - counterClockwiseMEA;
	}
	
	public String toString() {
		return "MEA[" + Logger.format(counterClockwiseMEA) + " to " + Logger.format(clockwiseMEA) + "]";
	}
}
