package xander.gfws.data;

import xander.gfws.distributer.TriangleDistributer;
import xander.gfws.distributer.WaveDistributer;
import xander.gfws.distributer.WeightDistributer;
import xander.gfws.modifier.FactorArrayModifier;
import xander.gfws.modifier.HeadOnFactorArrayModifier;
import xander.gfws.modifier.LinearFactorArrayModifier;

/**
 * KNNFactorArrayprocessor parameters used to determine how to build a factor array
 * from a given set of data provided by a LogReader.
 * 
 * @author Scott Arnold
 */
public class FactorArrayParms {
	
	WeightDistributer distributer;
	double weight = 3d;
	FactorArrayModifier arrayInitializer;
	int maxDataPointsForUsingArrayInitializer = 0; // by default, use initializer when less than 2 data points
	boolean useFlattener = false;
	double useFlattenerOnHitPercentage;
	double useFlattenerVisitPercentage;
	int useFlattenerEarliestRound;
	int flattenerOnCount;
	int flattenerOffCount;
	boolean useBulletShadows = true;
	boolean varyWeightByAge = false;
	boolean varyWeightByDistance = false;
	double decayPerTick = 0;   // in range 0 - 1
	boolean useCreativeAgeRoll = false;
	double minAgeWeightPercent = 0.01;
	double varyWeightByRankMultiplier = 1;
	
	public FactorArrayParms(boolean offensive) {
		if (offensive) {
			this.distributer = new TriangleDistributer();
			this.arrayInitializer = new LinearFactorArrayModifier(distributer, weight);
		} else {
			this.distributer = new WaveDistributer();
			this.arrayInitializer = new HeadOnFactorArrayModifier(distributer, weight);
		}
	}
	
	public WeightDistributer getDistributer() {
		return distributer;
	}

	public void setDistributer(WeightDistributer distributer) {
		this.distributer = distributer;
	}

	public void setMinAgeWeightPercent(double minAgeWeightPercent) {
		this.minAgeWeightPercent = minAgeWeightPercent;
	}

	public void setUseFlattener(boolean useFlattener, double opponentHitPercentage, 
			double visitPercentage, int earliestRound) {
		this.useFlattener = useFlattener;
		this.useFlattenerOnHitPercentage = opponentHitPercentage;
		this.useFlattenerEarliestRound = earliestRound;
		this.useFlattenerVisitPercentage = visitPercentage;
	}
	
	public void setUseBulletShadows(boolean useBulletShadows) {
		this.useBulletShadows = useBulletShadows;
	}
	
	public void setVaryWeightByAge(double decayPerTick) {
		this.decayPerTick = decayPerTick;
		this.varyWeightByAge = decayPerTick > 0;
	}

	public boolean isVaryWeightByDistance() {
		return varyWeightByDistance;
	}

	public void setVaryWeightByDistance(boolean varyWeightByDistance) {
		this.varyWeightByDistance = varyWeightByDistance;
	}

	public void setUseCreativeAgeRoll(boolean useCreativeAgeRoll) {
		this.useCreativeAgeRoll = useCreativeAgeRoll;
	}
	
	public void setArrayInitializer(FactorArrayModifier arrayInitializer) {
		this.arrayInitializer = arrayInitializer;
	}

	public void setMaxDataPointsForUsingArrayInitializer(
			int maxDataPointsForUsingArrayInitializer) {
		this.maxDataPointsForUsingArrayInitializer = maxDataPointsForUsingArrayInitializer;
	}

	public double getVaryWeightByRankMultiplier() {
		return varyWeightByRankMultiplier;
	}

	public void setVaryWeightByRankMultiplier(double varyWeightByRankMultiplier) {
		this.varyWeightByRankMultiplier = varyWeightByRankMultiplier;
	}
	
	
}
