package xander.gfws.segment;

import xander.core.log.Log;
import xander.core.log.Logger;
import xander.core.math.LinearEquation;
import xander.core.track.Wave;

public abstract class AbstractSegmenter implements Segmenter {

	private static final Log log = Logger.getLog(AbstractSegmenter.class);
	
	private LinearEquation le;
	private int numSegments;
	private double minOfRange;
	private double maxOfRange;
	private double range;
	
	public AbstractSegmenter(int numSegments, double minOfRange, double maxOfRange) {
		this.numSegments = numSegments;
		this.minOfRange = minOfRange;
		this.maxOfRange = maxOfRange;
		this.range = maxOfRange - minOfRange;
		this.le = new LinearEquation(minOfRange, 0, maxOfRange, numSegments, 0, numSegments-0.01);
	}

	protected abstract String getDescriptionForRange(int segmentIndex, double rangeStart, double rangeStop);
	
	@Override
	public int getNumSegments() {
		return numSegments;
	}

	@Override
	public int getSegmentIndex(Wave wave) {
		double doubleIndex = le.getY(getSegmentValue(wave));
		return (int) Math.round(Math.floor(doubleIndex));
	}

	@Override
	public double getSegmentValuePercent(Wave wave) {
		double segmentValue = Math.max(minOfRange, Math.min(maxOfRange, getSegmentValue(wave)));
		
		double svp = (segmentValue - minOfRange)/range;
		if (svp < 0 || svp > 1) {
			log.error("Segmentation percent svp out of range 0 to 1: svp=" + Logger.format(svp,2) + ";segmentValue=" + Logger.format(segmentValue,2) + ";minOfRange=" + Logger.format(minOfRange,2) + ";range=" + Logger.format(range,2));
		}
		return svp;
	}
	
	@Override
	public String getSegmentDescription(int segmentIndex) {
		double rangeStart = le.getX(segmentIndex);
		double rangeStop = le.getX(segmentIndex+1);
		return getDescriptionForRange(segmentIndex, rangeStart, rangeStop);
	}
}
