package xander.gfws.segment;

import xander.core.Resources;
import xander.core.track.Snapshot;
import xander.core.track.SnapshotHistory;
import xander.core.track.Wave;

public class DefenderAccelerationSegmenter implements Segmenter {

	private String myRobotName;
	
	public DefenderAccelerationSegmenter() {
		this.myRobotName = Resources.getRobotProxy().getName();
	}
	
	@Override
	public String getName() {
		return "Defender Acceleration Segmenter";
	}

	@Override
	public int getNumSegments() {
		return 3;
	}

	@Override
	public int getSegmentIndex(Wave wave) {
		SnapshotHistory history = Resources.getSnapshotHistory();
		Snapshot defenderSnapshot = wave.getInitialDefenderSnapshot();
		Snapshot previousDefenderSnapshot = null;
		if (defenderSnapshot.getName().equals(myRobotName)) {
			previousDefenderSnapshot = history.getMySnapshot(defenderSnapshot.getTime()-1);
		} else {
			previousDefenderSnapshot = history.getSnapshot(
					defenderSnapshot.getName(), defenderSnapshot.getTime()-1);
		}
		double s = defenderSnapshot.getVelocity();
		double ps = (previousDefenderSnapshot == null)? s : previousDefenderSnapshot.getVelocity();
		if (s == ps) {
			// steady speed
			return 0;			
		} else if ((s > 0 && ps < 0) || (s < 0 && ps > 0) || (Math.abs(s) > Math.abs(ps))) {
			// accelerating
			return 2;
		} else {
			// decelerating
			return 1;
		}
	}

	@Override
	public double getSegmentValue(Wave wave) {
		return getSegmentIndex(wave);
	}

	@Override
	public double getSegmentValuePercent(Wave wave) {
		int segmentIndex = getSegmentIndex(wave);
		switch (segmentIndex) {
		case 0:
			return 0.5;
		case 1:
			return 0;
		default:
			return 1;
		}
	}

	@Override
	public String getSegmentDescription(int segmentIndex) {
		switch (segmentIndex) {
		case 0:
			return "Steady Speed";
		case 1:
			return "Decelerating";
		default: 
			return "Accelerating";
		}
	}

}
