package xander.gfws.segment;

import xander.core.track.Wave;

public interface Segmenter {

	/**
	 * Returns a short name for this segmenter.
	 * 
	 * @return     name for segmenter
	 */
	public String getName();
	
	/**
	 * Returns the total number of segments this segmenter uses.
	 * 
	 * @return     total number of segments this segmenter uses.
	 */
	public int getNumSegments();
	
	/**
	 * Returns the segment index that the given opponent should fall into.
	 * The segment index should range from 0 to the number of segments less 1.
	 * 
	 * @param bulletWave    bullet wave to get segment index for
	 * 
	 * @return        segment index
	 */
	public int getSegmentIndex(Wave wave);
	
	/**
	 * Returns the value used to determine where the wave falls in the segment.
	 * This value can be used in place of the segment index for systems that
	 * handle segment slices in a dynamic manner.
	 * 
	 * @param wave       wave
	 * 
	 * @return        segment value
	 */
	public double getSegmentValue(Wave wave);
	
	/**
	 * Returns a normalized percent-in-range for the segment value.  For
	 * example, if the segment range is 0 to 10, and the segment value is
	 * 4, this method should return a value of 0.4)
	 * 
	 * @param wave      wave
	 * 
	 * @return          normalized percent-in-range for the segment value
	 */
	public double getSegmentValuePercent(Wave wave);
	
	/**
	 * Returns a description of the segment for the given segment index.
	 * 
	 * @param segmentIndex       index of segment
	 * 
	 * @return        description of segment
	 */
	public String getSegmentDescription(int segmentIndex);
}
